"""Instrument driver for the PhysikInstrumente E-873 Servo Controller."""

import enum
import logging
import re
import time
from typing import List, Mapping, NamedTuple, Optional, Tuple

from qmi.core.context import QMI_Context
from qmi.core.exceptions import QMI_TimeoutException, QMI_InstrumentException
from qmi.core.instrument import QMI_Instrument, QMI_InstrumentIdentification
from qmi.core.rpc import rpc_method
from qmi.core.transport import create_transport

from .gcs_error_codes import GCS_CONTROLLER_ERROR_CODES


# Global variable holding the logger for this module.
_logger = logging.getLogger(__name__)


class ReferenceTarget(enum.Enum):
    """Target position of reference move."""
    REFERENCE_POINT = 0
    POSITIVE_LIMIT = 1
    NEGATIVE_LIMIT = -1


class ReferenceSignalMode(enum.Enum):
    """Type of reference signal/approach."""
    DIRECTION_SENSING = 0
    PULSE_SIGNAL = 1
    INDEX_PULSE_VIA_NEGATIVE_LIMIT = 2
    INDEX_PULSE_VIA_POSITIVE_LIMIT = 3


# Named tuple returned by get_stage_info() command.
StageInfo = NamedTuple('StageInfo', [
    ('type', str),
    ('serial_number', str),
    ('assembly_date', str),
    ('hw_version', str)
])

# Named tuple returned by get_system_status() command.
SystemStatus = NamedTuple('SystemStatus', [
    ('negative_limit_switch', bool),
    ('reference_point_switch', bool),
    ('positive_limit_switch', bool),
    ('digital_input', Tuple[bool, bool, bool, bool]),
    ('error_flag', bool),
    ('servo_mode', bool),
    ('in_motion', bool),
    ('on_target', bool)
])


class PI_E873(QMI_Instrument):
    """Instrument driver for the PhysikInstrumente E-873 Servo Controller."""

    _rpc_constants = ["ERROR_CODES"]

    # Error codes returned by instrument via ERR? query.
    ERROR_CODES = GCS_CONTROLLER_ERROR_CODES

    # Instrument should respond within 2 seconds.
    RESPONSE_TIMEOUT = 2.0

    def __init__(self, context: QMI_Context, name: str, transport: str) -> None:
        """Initialize driver.

        :argument name: Name for this instrument instance.
        :argument transport: QMI transport descriptor for command interface.
        """
        super().__init__(context, name)
        self.axis_id = "1"
        self._transport = create_transport(transport)

    @rpc_method
    def open(self) -> None:
        _logger.info("Opening connection to %s", self._name)
        self._transport.open()
        super().open()
        # Clear the error register.
        self._ask("ERR?")
        # Check axis identifiers. There should be a single axis, and the axis ID should be "1".
        resp = self._ask("SAI?")
        axis_ids = [t.strip() for t in resp.split("\n")]
        if axis_ids != [self.axis_id]:
            raise QMI_InstrumentException("Unexpected axis identifiers {!r}".format(axis_ids))

    @rpc_method
    def close(self) -> None:
        _logger.info("Closing connection to %s", self._name)
        super().close()
        self._transport.close()

    def _write(self, cmd: str) -> None:
        """Send command to instrument and check instrument responds with "OK"."""
        self._check_is_open()

        # There are a few special one-character commands which must not
        # be terminated. Other commands are terminated with \n.
        rawcmd = cmd.encode('ascii')
        if len(rawcmd) != 1:
            rawcmd += b"\n"

        # Send command.
        self._transport.write(rawcmd)

    def _ask(self, cmd: str) -> str:
        """Send command to instrument and return response from instrument."""
        self._check_is_open()

        # Discard pending data received from the controller.
        # Normally there should not be any pending data. However if a
        # previous query timed out, the belated response to the previous
        # query may now be sitting in the receive buffer. In that case we
        # have to discard it, otherwise we will be forever out of sync
        # with the controller.
        self._transport.discard_read()

        # There are a few special one-character commands which must not
        # be terminated. Other commands are terminated with \n.
        rawcmd = cmd.encode('ascii')
        if len(rawcmd) != 1:
            rawcmd += b"\n"

        # Send command.
        self._transport.write(rawcmd)

        # Read response.
        # NOTE: The instrument can send a multi-line responses. In that case
        # every line except the last line ends with a space character.
        resp = bytearray()
        lines = []  # type: List[bytes]
        while True:
            line = self._transport.read_until(message_terminator=b"\n",
                                              timeout=self.RESPONSE_TIMEOUT)
            if not line and not resp:
                raise QMI_TimeoutException("No response to command {!r}".format(cmd))
            if not line.endswith(b"\n"):
                raise QMI_InstrumentException("Got partial response to command {!r}: {!r}".format(cmd, line))
            line = line.rstrip(b"\r\n")
            if line.endswith(b" "):
                # Multi-line response, to be continued.
                resp.extend(line[:-1])
                resp.extend(b"\n")
            else:
                # This was the last line of the response.
                resp.extend(line)
                break

        return resp.decode('ascii', errors='replace')

    @staticmethod
    def _parse_response_item(resp: str, cmd: str, key: str) -> str:
        """Extract value part from a single-line "key=value" response from the instrument."""
        lines = resp.splitlines()
        if len(lines) != 1:
            raise QMI_InstrumentException(
                "Expecting single line response to command {!r} but got {!r}".format(cmd, resp))
        if not resp.startswith(key + "="):
            raise QMI_InstrumentException(
                "Expecting response with key {!r} to command {!r} but got {!r}".format(key, cmd, resp))
        return resp[len(key) + 1:]

    @staticmethod
    def _parse_int(resp: str, cmd: str, key: str = "") -> int:
        """Parse integer response from instrument.

        In some cases, the instrument response is formatted as "key=value".
        This function will parse that format if the expected key is specified.

        :argument resp: Response string from instrument.
        :argument cmd: Command string (for error messages).
        :argument key: Optional expected key for responses in key=value format.
        :return: Integer value.
        :raises QMI_InstrumentException: In case of unexpected response format.
        """
        if key:
            # Extract value part from "key=value" response.
            resp = PI_E873._parse_response_item(resp, cmd, key)
        try:
            return int(resp, 0)  # must also support hexadecimal format
        except ValueError:
            raise QMI_InstrumentException("Expecting integer response to command {!r} but got {!r}".format(cmd, resp))

    @staticmethod
    def _parse_float(resp: str, cmd: str, key: str = "") -> float:
        """Parse floating point response from instrument.

        :argument resp: Response string from instrument.
        :argument cmd: Command string (for error messages).
        :argument key: Optional expected key for responses in key=value format.
        :return: Floating point value.
        :raises QMI_InstrumentException: In case of unexpected response format.
        """
        if key:
            # Extract value part from "key=value" response.
            resp = PI_E873._parse_response_item(resp, cmd, key)
        try:
            return float(resp)
        except ValueError:
            raise QMI_InstrumentException("Expecting integer response to command {!r} but got {!r}".format(cmd, resp))

    def _check_error(self, expect_stop: bool = False) -> None:
        """Read error code from the instrument and raise exception in case of error."""
        cmd = "ERR?"
        resp = self._ask(cmd)
        err = self._parse_int(resp, cmd)
        # Ignore STOP error following a STP or HLT command.
        if err != 0 and ((not expect_stop) or err != 10):
            err_str = self.ERROR_CODES.get(err, "unknown")
            raise QMI_InstrumentException("Instrument reports error {} ({})".format(err, err_str))

    def _get_float_param(self, item: str, param: int) -> float:
        """Read floating point parameter from controller."""
        key = "{} 0X{:X}".format(item, param)
        cmd = "SPA? {}".format(key)
        resp = self._ask(cmd)
        return self._parse_float(resp, cmd, key=key)

    def _get_char_param(self, item: str, param: int) -> str:
        """Read string parameter from controller."""
        key = "{} 0X{:X}".format(item, param)
        cmd = "SPA? {}".format(key)
        resp = self._ask(cmd)
        return self._parse_response_item(resp, cmd, key=key)

    @rpc_method
    def reset(self) -> None:
        """Reboot the instrument, returning (most) settings to their defaults."""
        _logger.info("[%s] reboot", self._name)
        self._write("RBT")
        # Delay until controller is ready to handle new commands.
        # If a new command is sent too quickly, the controller may stop responding.
        time.sleep(1.0)
        self._check_error()

    @rpc_method
    def get_idn(self) -> QMI_InstrumentIdentification:
        """Read instrument type and version and return QMI_InstrumentIdentification instance."""
        cmd = "*IDN?"
        resp = self._ask(cmd)
        words = resp.rstrip().split(",")
        if len(words) != 4:
            raise QMI_InstrumentException("Unexpected response to {!r}, got {!r}".format(cmd, resp))
        return QMI_InstrumentIdentification(vendor=words[0].strip(),
                                            model=words[1].strip(),
                                            serial=words[2].strip(),
                                            version=words[3].strip())

    @rpc_method
    def get_stage_info(self) -> StageInfo:
        """Read information about the connected motor stage."""
        stage_type = self._get_char_param(self.axis_id, 0x0f000100)
        stage_serial = self._get_char_param(self.axis_id, 0x0f000200)
        stage_date = self._get_char_param(self.axis_id, 0x0f000300)
        stage_version = self._get_char_param(self.axis_id, 0x0f000400)
        return StageInfo(type=stage_type,
                         serial_number=stage_serial,
                         assembly_date=stage_date,
                         hw_version=stage_version)

    @rpc_method
    def get_error(self) -> int:
        """Return last error code and clear error register.

        Note most commands functions in this driver will automatically check
        the error register after sending the command.

        :return: Error core, or 0 if no error occurred since last command.
        """
        cmd = "ERR?"
        resp = self._ask(cmd)
        return self._parse_int(resp, cmd)

    @rpc_method
    def get_system_status(self) -> SystemStatus:
        """Request system status information.

        :return: 16-bit status code
        """
        cmd = "\x04"  # shortcut for "SRG?"
        resp = self._ask(cmd)
        w = self._parse_int(resp, cmd)
        # Decode status word.
        return SystemStatus(negative_limit_switch=((w & 1) != 0),
                            reference_point_switch=((w & 2) != 0),
                            positive_limit_switch=((w & 4) != 0),
                            digital_input=(((w & 0x10) != 0),
                                           ((w & 0x20) != 0),
                                           ((w & 0x40) != 0),
                                           ((w & 0x80) != 0)),
                            error_flag=((w & 0x100) != 0),
                            servo_mode=((w & 0x1000) != 0),
                            in_motion=((w & 0x2000) != 0),
                            on_target=((w & 0x8000) != 0))

    @rpc_method
    def stop_all(self) -> None:
        """Stop motion abruptly and stop macro execution."""
        _logger.debug("[%s] stop all axes", self._name)
        cmd = "\x18"  # shortcut for "STP"
        self._write(cmd)
        self._check_error(expect_stop=True)

    @rpc_method
    def stop_smooth(self) -> None:
        """Stop motion smoothly (taking deceleration into account)."""
        _logger.debug("[%s] halt axis", self._name)
        self._write("HLT")
        self._check_error(expect_stop=True)

    @rpc_method
    def get_physical_unit(self) -> str:
        """Return physical unit used for axis positions."""
        return self._get_char_param(self.axis_id, 0x07000601)

    @rpc_method
    def get_position_range(self) -> Tuple[float, float]:
        """Return range of commandable positions.

        :return: Range as tuple (min_pos, max_pos) in physical units.
        """
        cmd = "TMN? {}".format(self.axis_id)
        resp = self._ask(cmd)
        pmin = self._parse_float(resp, cmd, key=self.axis_id)
        cmd = "TMX? {}".format(self.axis_id)
        resp = self._ask(cmd)
        pmax = self._parse_float(resp, cmd, key=self.axis_id)
        return (pmin, pmax)

    @rpc_method
    def get_max_position_error(self) -> float:
        """Return the maximum allowed position error before the controller stops."""
        return self._get_float_param(self.axis_id, 0x8)

    @rpc_method
    def get_max_velocity(self) -> float:
        """Return the maximum closed-loop velocity in phys_unit/second."""
        return self._get_float_param(self.axis_id, 0xa)

    @rpc_method
    def get_max_acceleration(self) -> float:
        """Return the maximum closed-loop acceleration in phys_unit/second/second."""
        return self._get_float_param(self.axis_id, 0x4a)

    @rpc_method
    def get_max_deceleration(self) -> float:
        """Return the maximum closed-loop deceleration in phys_unit/second/second."""
        return self._get_float_param(self.axis_id, 0x4b)

    @rpc_method
    def get_reference_signal_mode(self) -> ReferenceSignalMode:
        """Return the method used by the motor controller to find the reference point."""
        v = self._get_float_param(self.axis_id, 0x70)
        return ReferenceSignalMode(int(v))

    @rpc_method
    def get_reference_velocity(self) -> float:
        """Return the velocity used for reference moves."""
        return self._get_float_param(self.axis_id, 0x50)

    @rpc_method
    def set_acceleration(self, accel: float) -> None:
        """Set acceleration for closed-loop motion.

        :argument accel: Acceleration in phys_unit/second/second.
        """
        _logger.debug("[%s] set acceleration %f", self._name, accel)
        self._write("ACC {} {:.8f}".format(self.axis_id, accel))
        self._check_error()

    @rpc_method
    def get_acceleration(self) -> float:
        """Get acceleration for closed-loop motion.

        :return: Acceleration in phys_unit/second/second.
        """
        cmd = "ACC? {}".format(self.axis_id)
        resp = self._ask(cmd)
        return self._parse_float(resp, cmd, key=self.axis_id)

    @rpc_method
    def set_deceleration(self, decel: float) -> None:
        """Set deceleration for closed-loop motion.

        :argument decel: Deceleration in phys_unit/second/second.
        """
        _logger.debug("[%s] set deceleration %f", self._name, decel)
        self._write("DEC {} {:.8f}".format(self.axis_id, decel))
        self._check_error()

    @rpc_method
    def get_deceleration(self) -> float:
        """Get deceleration for closed-loop motion.

        :return: Deceleration in phys_unit/second/second.
        """
        cmd = "DEC? {}".format(self.axis_id)
        resp = self._ask(cmd)
        return self._parse_float(resp, cmd, key=self.axis_id)

    @rpc_method
    def set_velocity(self, velocity: float) -> None:
        """Set velocity for closed-loop motion.

        :argument velocity: Velocity is phys_unit/second.
        """
        _logger.debug("[%s] set velocity %f", self._name, velocity)
        if velocity < 0:
            raise QMI_InstrumentException("Invalid velocity {}".format(velocity))
        self._write("VEL {} {:.8f}".format(self.axis_id, velocity))
        self._check_error()

    @rpc_method
    def get_velocity(self) -> float:
        """Get velocity for closed-loop motion.

        :return: Velocity in phys_unit/second.
        """
        cmd = "VEL? {}".format(self.axis_id)
        resp = self._ask(cmd)
        return self._parse_float(resp, cmd, key=self.axis_id)

    @rpc_method
    def set_servo_mode(self, enable: bool) -> None:
        """Enable or disable closed-loop servo operation.

        Servo mode must be enabled for reference moves and for closed-loop positioning.
        """
        _logger.debug("[%s] set servo mode %d", self._name, enable)
        state = 1 if enable else 0
        self._write("SVO {} {}".format(self.axis_id, state))
        self._check_error()

    @rpc_method
    def get_servo_mode(self) -> bool:
        """Return True if closed-loop servo mode is active."""
        cmd = "SVO? {}".format(self.axis_id)
        resp = self._ask(cmd)
        v = self._parse_int(resp, cmd, key=self.axis_id)
        return v == 1

    @rpc_method
    def set_reference_definition_mode(self, force_reference: bool) -> None:
        """Set reference point definition mode.

        :argument force_reference: True if a reference move is required for
            closed-loop operation; False if relative moves are allowed on
            an unreferenced axis.
        """
        _logger.debug("[%s] set reference definition mode %d", self._name, force_reference)
        mode = "1" if force_reference else "0"
        self._write("RON {} {}".format(self.axis_id, mode))
        self._check_error()

    @rpc_method
    def get_reference_definition_mode(self) -> bool:
        """Return reference point definition mode.

        :return: True if a reference move is required for closed-loop operation;
            False if relative moves are allowed on an unreferenced axis.
        """
        cmd = "RON? {}".format(self.axis_id)
        resp = self._ask(cmd)
        v = self._parse_int(resp, cmd, key=self.axis_id)
        return v == 1

    @rpc_method
    def reference_move(self,
                       target: ReferenceTarget = ReferenceTarget.REFERENCE_POINT) -> None:
        """Start a reference move to a well-defined position.

        When the axis reaches the target position, the absolute position
        of the axis is reset to the known position of the target.
        From that moment on, the axis is referenced and absolute closed-loop
        motion is possible.

        Servo mode must be enabled before calling this function.
        See set_servo_mode().

        Motion of the axis may continue after this function returns.
        Call get_system_status() to determine when motion is complete.
        Call stop_smooth() to abort the reference move.

        :argument target: Target of reference move.
            REFERENCE_POINT means move to the positive or negative axis
            limit (depending on stage parameters), then move back to the
            reference point.
            POSITIVE_LIMIT means move to the positive axis limit.
            NEGATIVE_LIMIT means move to the negative axis limit.
        """
        _logger.debug("[%s] start reference move", self._name)
        if target == ReferenceTarget.POSITIVE_LIMIT:
            self._write("FPL")
        elif target == ReferenceTarget.NEGATIVE_LIMIT:
            self._write("FNL")
        else:
            self._write("FRF")
        self._check_error()

    @rpc_method
    def get_reference_result(self) -> bool:
        """Return True if the axis is referenced (absolute position is known).

        :return: True if axis is referenced, False if axis is not referenced.
        """
        cmd = "FRF? {}".format(self.axis_id)
        resp = self._ask(cmd)
        v = self._parse_int(resp, cmd, key=self.axis_id)
        return v == 1

    @rpc_method
    def move_absolute(self, position: float) -> None:
        """Start closed-loop motion to absolute position.

        The axis must be referenced before absolute motion is possible.
        See reference_move().

        Motion of the axis may continue after this function returns.
        Call get_system_status() to determine when motion is complete.
        Call stop_smooth() to abort the reference move.

        :argument position: Target position in physical units.
        """
        _logger.debug("[%s] absolute move to %f", self._name, position)
        self._write("MOV {} {:.8f}".format(self.axis_id, position))
        self._check_error()

    @rpc_method
    def move_relative(self, displacement: float) -> None:
        """Start closed-loop motion to a target relative to the current position.

        Before calling this function, servo mode must be enabled and either
        the axis must be referenced or using an unreferenced axis must be
        allowed. See set_reference_definition_mode().

        Motion of the axis may continue after this function returns.
        Call get_system_status() to determine when motion is complete.
        Call stop_smooth() to abort the reference move.

        :argument displacement: Displacement (in physical units) from current
            position to target.
        """
        _logger.debug("[%s] relative move by %f", self._name, displacement)
        self._write("MVR {} {:.8f}".format(self.axis_id, displacement))
        self._check_error()

    @rpc_method
    def get_target_position(self) -> float:
        """Get target position of axis.

        :return: Target position in physical units.
        """
        cmd = "MOV? {}".format(self.axis_id)
        resp = self._ask(cmd)
        return self._parse_float(resp, cmd, key=self.axis_id)

    @rpc_method
    def get_position(self) -> float:
        """Get current absolute position of axis.

        :return: Current position in physical units.
        """
        cmd = "POS? {}".format(self.axis_id)
        resp = self._ask(cmd)
        return self._parse_float(resp, cmd, key=self.axis_id)

    @rpc_method
    def wait_motion_complete(self, timeout: Optional[float] = None) -> bool:
        """Wait until the current motion ends.

        :argument timeout: Maximum time to wait in seconds, or None to wait indefinitely.
        :return: True if motion complete, False if timeout expired before motion complete.
        """
        _logger.debug("[%s] wait until motion complete", self._name)

        endtime = None  # type: Optional[float]
        if timeout is not None:
            endtime = time.monotonic() + timeout

        while True:
            status = self.get_system_status()
            if not status.in_motion:
                return True
            if endtime is not None:
                t = time.monotonic()
                if t >= endtime:
                    return False
            time.sleep(0.02)

    @rpc_method
    def set_trigger_inmotion(self, dig_out: int) -> None:
        """Set digital output to be active as the axis is in motion.

        :argument dig_out: Digital output line (range 1 to 4)
        """

        if dig_out not in (1, 2, 3, 4):
            raise QMI_InstrumentException("Invalid digital output line {}".format(dig_out))

        _logger.debug("[%s] set trigger InMotion dig_out=%d", self._name, dig_out)

        # Set axis 1.
        self._write("CTO {} 2 {}".format(dig_out, self.axis_id))

        # Set trigger mode InMotion.
        self._write("CTO {} 3 6".format(dig_out))
        self._check_error()

    @rpc_method
    def set_trigger_position_offset(self, dig_out: int, step: float, start: float, stop: float) -> None:
        """Set digital output to pulse at specified position intervals.

        The first pulse is triggered when the axis has reached the specified offset position.
        Subsequent pulses are triggered each time the axis position equals the sum of the
        last trigger position and the specified step size. Trigger output ends when the
        axis position exceeds the specified stop position.

        :argument dig_out: Digital output line (range 1 to 4).
        :argument step: Position increment (physical units) between output pulses.
            This parameter must be negative if the axis moves in negative direction.
        :argument start: Position (physical units) of first output pulse..
        :argument stop: Position (physical units) where the output pulse stops.
        """

        if dig_out not in (1, 2, 3, 4):
            raise QMI_InstrumentException("Invalid digital output line {}".format(dig_out))

        _logger.debug("[%s] set trigger Position+Offset dig_out=%d", self._name, dig_out)

        # Set axis 1.
        self._write("CTO {} 2 {}".format(dig_out, self.axis_id))

        # Set TriggerStep.
        self._write("CTO {} 1 {:.8f}".format(dig_out, step))

        # Set TriggerPosition.
        self._write("CTO {} 10 {:.8f}".format(dig_out, start))

        # Set StopThreshold.
        self._write("CTO {} 9 {:.8f}".format(dig_out, stop))

        # Set trigger mode Position+Offset
        self._write("CTO {} 3 7".format(dig_out))
        self._check_error()

    @rpc_method
    def set_trigger_output_state(self, dig_out: int, enable: bool) -> None:
        """Enable or disable trigger output on the specified digital output line."""
        if dig_out not in (1, 2, 3, 4):
            raise QMI_InstrumentException("Invalid digital output line {}".format(dig_out))
        _logger.debug("[%s] set trigger state dig_out=%d enable=%d", self._name, dig_out, enable)
        mode = 1 if enable else 0
        self._write("TRO {} {}".format(dig_out, mode))
        self._check_error()

    @rpc_method
    def get_trigger_output_state(self, dig_out: int) -> bool:
        """Return True if trigger output is enabled on the specified digital output line."""
        if dig_out not in (1, 2, 3, 4):
            raise QMI_InstrumentException("Invalid digital output line {}".format(dig_out))
        cmd = "TRO? {}".format(dig_out)
        resp = self._ask(cmd)
        v = self._parse_int(resp, cmd, key=str(dig_out))
        return v == 1

    @rpc_method
    def set_digital_output(self, dig_out: int, state: bool) -> None:
        """Switch the specified digital output line to the specified state.

        This command must not be used on digital lines for which
        the trigger output is enabled.

        :argument dig_out: Digital output line (range 1 to 4).
        :argument state: True to set line high, False to set line low.
        """
        if dig_out not in (1, 2, 3, 4):
            raise QMI_InstrumentException("Invalid digital output line {}".format(dig_out))
        _logger.debug("[%s] set digital output %d = %d", self._name, dig_out, state)
        v = 1 if state else 0
        self._write("DIO {} {}".format(dig_out, v))
        self._check_error()

    @rpc_method
    def get_digital_input(self, dig_in: int) -> bool:
        """Read the state of the specified digital input line.

        :argument dig_in: Digital input line (range 1 to 4).
        :return: True if the input signal is high, False if the signal is low.
        """
        if dig_in not in (1, 2, 3, 4):
            raise QMI_InstrumentException("Invalid digital input line {}".format(dig_in))
        cmd = "DIO? {}".format(dig_in)
        resp = self._ask(cmd)
        v = self._parse_int(resp, cmd, key=str(dig_in))
        return v == 1

    @rpc_method
    def define_macro(self, name: str, commands: List[str]) -> None:
        """Define a new macro.

        :argument name: Name of the new macro.
        :argument commands: List of command strings to put in the macro.
        """
        cmd = "MAC BEG {}".format(name)
        self._write(cmd)
        for cmd in commands:
            time.sleep(0.01)  # avoid overflowing the controller input buffer
            self._write(cmd)
        cmd = "MAC END"
        self._write(cmd)
        self._check_error()

    @rpc_method
    def delete_macro(self, name: str) -> None:
        """Delete the specified macro."""
        self._write("MAC DEL {}".format(name))
        self._check_error()

    @rpc_method
    def get_defined_macros(self) -> List[str]:
        """Return a list of names of all defined macros."""
        cmd = "MAC?"
        resp = self._ask(cmd)
        return resp.splitlines()

    @rpc_method
    def start_macro(self,
                    name: str,
                    arg1: Optional[str] = None,
                    arg2: Optional[str] = None,
                    repeat: int = 1) -> None:
        """Start the specified macro one or more times.

        This function returns immediately after starting the macro.
        While the macro executes, the controller can process other commands.
        Only one macro can run at a time.

        If any command executed by the macro causes an error,
        the error status will appear in the global error register.
        A subsequent direct command to the controller may then
        report the pending error which was caused by the macro.
        To avoid this, call get_error() after macro execution completes
        to reset the global error register.

        Call stop_all() to abort all running macros.

        :argument arg1: Optional argument to load into VAR 1.
        :argument arg2: Optional argument to load into VAR 2.
        :argument repeat: Number of executions of the macro (default 1).
        """
        if (arg1 is None) and (arg2 is not None):
            raise QMI_InstrumentException("Can not call macro without arg1 but with arg2")
        if repeat < 1:
            raise QMI_InstrumentException("Can not execute macro fewer than 1 times")
        args = ""
        if arg1 is not None:
            args += " " + arg1
        if arg2 is not None:
            args += " " + arg2
        if repeat == 1:
            cmd = "MAC START {}{}".format(name, args)
        else:
            cmd = "MAC NSTART {} {}{}".format(name, repeat, args)
        self._write(cmd)
        self._check_error()

    @rpc_method
    def get_running_macros(self) -> List[str]:
        """Return a list of names of currently running macros."""
        cmd = "RMC?"
        resp = self._ask(cmd)
        return resp.splitlines()

    @rpc_method
    def set_variable(self, name: str, value: Optional[str]) -> None:
        """Set the value of a global variable.

        :argument name: Name of the variable.
        :argument value: New value for the variable, or None to delete the variable.
        """
        if not re.match(r"^[a-zA-Z][a-zA-Z0-9]{0,7}$", name):
            raise ValueError("Invalid variable name {!r}".format(name))
        if value is not None:
            cmd = "VAR {} {}".format(name, value)
        else:
            cmd = "VAR {}".format(name)
        self._write(cmd)
        self._check_error()

    @rpc_method
    def get_variable(self, name: str) -> str:
        """Get the value of a global variable.

        :argument name: Name of the variable.
        :return: Current value of the variable.
        """
        if not re.match(r"^[a-zA-Z][a-zA-Z0-9]{0,7}$", name):
            raise ValueError("Invalid variable name {!r}".format(name))
        cmd = "VAR? {}".format(name)
        resp = self._ask(cmd)
        return self._parse_response_item(resp, cmd, key=name)

    @rpc_method
    def get_variables(self) -> Mapping[str, str]:
        """Get all global variables.

        :return: Dictionary of name, value pairs.
        """
        cmd = "VAR?"
        resp = self._ask(cmd)
        ret = {}
        for line in resp.splitlines():
            p = line.find("=")
            if p < 0:
                raise QMI_InstrumentException(
                    "Expecting response format key=value but got {!r}".format(line))
            ret[line[:p]] = line[p + 1:]
        return ret



"""
Instrument driver for the Stanford Research Systems DC205 voltage source.
"""

import logging
from typing import List

from qmi.core.context import QMI_Context
from qmi.core.exceptions import QMI_InstrumentException
from qmi.core.instrument import QMI_Instrument, QMI_InstrumentIdentification
from qmi.core.rpc import rpc_method
from qmi.core.scpi_protocol import ScpiProtocol
from qmi.core.transport import create_transport

# Global variable holding the logger for this module.
_logger = logging.getLogger(__name__)

# Execution error codes returned by the DC205.
EXECUTION_ERRORS = {
    0: "Ok",
    1: "Illegal value",
    2: "Wrong token",
    3: "Invalid bit",
    4: "Queue full",
    5: "Not compatible",
}

# Command error codes returned by the DC205.
COMMAND_ERRORS = {
    0: "Ok",
    1: "Illegal command",
    2: "Undefined command",
    3: "Illegal query",
    4: "Illegal set",
    5: "Missing parameter",
    6: "Extra parameter",
    7: "Null parameter",
    8: "Parameter buffer overflow",
    9: "Bad floating-point",
    10: "Bad integer",
    11: "Bad integer token",
    12: "Bad token value",
    13: "Bad hex block",
    14: "Unknown token",
}


class SRS_DC205(QMI_Instrument):
    """Instrument driver for the Stanford Research Systems DC205 voltage source.

    This driver implements most basic features of the instrument.
    Linear voltage scanning is not supported by this driver.
    """

    # Response timeout for normal commands.
    COMMAND_RESPONSE_TIMEOUT = 2.0

    # Response timeout for reset command.
    RESET_RESPONSE_TIMEOUT = 5.0

    def __init__(self, context: QMI_Context, name: str, transport: str) -> None:
        """Initialize the instrument driver.

        Arguments:
            name:       Name for this instrument instance.
            transport:  QMI transport descriptor to connect to the instrument.
                            The transport descriptor will typically specify a
                            serial port and baud rate, for example
                            "serial:/dev/ttyUSB0:baudrate=115200".
        """
        super().__init__(context, name)
        self._transport = create_transport(
            transport, default_attributes={"baudrate": 115200}
        )
        self._scpi_protocol = ScpiProtocol(
            self._transport,
            command_terminator="\n",
            response_terminator="\n",
            default_timeout=self.COMMAND_RESPONSE_TIMEOUT,
        )

    @rpc_method
    def open(self) -> None:
        _logger.info("[%s] Opening connection to instrument", self._name)
        self._transport.open()
        # The DC205 supports several different response termination styles.
        # Make sure that the instrument is configured for the style expected by the ScpiProtocol handler.
        try:
            self._transport.write(b"\nTERM LF\n")
        except Exception:
            self._transport.close()
            raise
        super().open()

    @rpc_method
    def close(self) -> None:
        _logger.info("[%s] Closing connection to instrument", self._name)
        super().close()
        self._transport.close()

    def _check_error(self, cmd: str) -> None:
        """Read the instrument error status and raise an exception if an error occurred."""
        resp = self._scpi_protocol.ask("LEXE?; LCME?")
        words = resp.split(";")
        if len(words) != 2:
            raise QMI_InstrumentException(
                f"Unexpected response to LEXE?;LCME?, got {resp!r}"
            )
        execution_error = int(words[0].strip())
        command_error = int(words[1].strip())
        if execution_error != 0:
            error_str = EXECUTION_ERRORS.get(execution_error, "unknown")
            raise QMI_InstrumentException(
                f"Instrument returned execution error after {cmd!r}: {execution_error} ({error_str})"
            )
        if command_error != 0:
            error_str = COMMAND_ERRORS.get(execution_error, "unknown")
            raise QMI_InstrumentException(
                f"Instrument returned command error after {cmd!r}: {command_error} ({error_str})"
            )

    def _set_command(self, cmd: str) -> None:
        self._check_is_open()
        self._scpi_protocol.write(cmd)
        self._check_error(cmd)

    def _ask_float(self, cmd: str) -> float:
        """Send a query and return a floating-point response."""
        self._check_is_open()
        resp = self._scpi_protocol.ask(cmd)
        try:
            return float(resp)
        except ValueError as exc:
            raise QMI_InstrumentException(
                f"Unexpected response to command {cmd!r}: {resp!r}"
            ) from exc

    def _ask_token(self, cmd: str, tokens: List[str]) -> int:
        """Read a token response from the instrument.

        The instrument can report a token value either as an integer
        or as a short string. This function accepts both forms and always
        returns the integer value.
        """
        self._check_is_open()
        resp = self._scpi_protocol.ask(cmd)
        resp = resp.strip()
        if resp in tokens:
            return tokens.index(resp)
        try:
            tok = int(resp)
        except ValueError as exc:
            raise QMI_InstrumentException(
                f"Unexpected response to command {cmd!r}: {resp!r}"
            ) from exc
        if tok >= len(tokens):
            raise QMI_InstrumentException(
                f"Unexpected response to command {cmd!r}: {resp!r}"
            )
        return tok

    @rpc_method
    def reset(self) -> None:
        """Reset the instrument, returning (most) settings to their defaults."""

        self._check_is_open()

        # Clear error registers.
        self._scpi_protocol.ask("LEXE?; LCME?", timeout=self.RESET_RESPONSE_TIMEOUT)

        # Reset instrument.
        self._scpi_protocol.write("*RST")

        # Wait until reset finished.
        self._scpi_protocol.ask("*OPC?", timeout=self.RESET_RESPONSE_TIMEOUT)

        # Check for errors occurred during reset.
        self._check_error("*RST")

    @rpc_method
    def get_idn(self) -> QMI_InstrumentIdentification:
        """Read instrument type and version and return QMI_InstrumentIdentification instance."""
        self._check_is_open()
        resp = self._scpi_protocol.ask("*IDN?")
        words = resp.rstrip().split(",")
        if len(words) != 4:
            raise QMI_InstrumentException(f"Unexpected response to *IDN?, got {resp!r}")
        return QMI_InstrumentIdentification(
            vendor=words[0].strip(),
            model=words[1].strip(),
            serial=words[2].strip(),
            version=words[3].strip(),
        )

    @rpc_method
    def get_range(self) -> int:
        """Read the output voltage range.

        The output range is always bipolar.
        Thus, range setting 1 means -1 .. +1 Volt; range setting 10 means -10 .. +10 Volt, etc.

        Returns:
            Output range in Volt. Possible return values are 1, 10 and 100.
        """
        tok = self._ask_token(cmd="RNGE?", tokens=["RANGE1", "RANGE10", "RANGE100"])
        return [1, 10, 100][tok]

    @rpc_method
    def set_range(self, volt_range: int) -> None:
        """Set the output voltage range.

        This command may not be used while the output is enabled.

        Parameters:
            volt_range: Output range in Volt. Allowed values are 1, 10 and 100.
        """
        if volt_range not in (1, 10, 100):
            raise ValueError("Invalid voltage range")
        self._set_command(f"RNGE RANGE{int(volt_range)}")

    @rpc_method
    def get_output_enabled(self) -> bool:
        """Return True if the output is enabled, False if the output is disabled."""
        tok = self._ask_token(cmd="SOUT?", tokens=["OFF", "ON"])
        return tok != 0

    @rpc_method
    def set_output_enabled(self, enable: bool) -> None:
        """Enable or disable the output."""
        self._set_command(f"SOUT {int(enable)}")

    @rpc_method
    def get_voltage(self) -> float:
        """Return the DC voltage setting in Volt.

        Note that this returns the voltage *setting* of the instrument.
        The instrument does not measure the *actual* voltage on its output.
        """
        return self._ask_float("VOLT?")

    @rpc_method
    def set_voltage(self, voltage: float) -> None:
        """Set the DC voltage setting in Volt."""
        self._set_command(f"VOLT {voltage:.6f}")

    @rpc_method
    def get_output_floating(self) -> bool:
        """Return True if floating output is enabled, False if output is ground-referenced."""
        tok = self._ask_token(cmd="ISOL?", tokens=["GROUND", "FLOAT"])
        return tok != 0

    @rpc_method
    def set_output_floating(self, enable: bool) -> None:
        """Enable or disable floating output."""
        self._set_command(f"ISOL {int(enable)}")

    @rpc_method
    def get_sensing_enabled(self) -> bool:
        """Return True if remote sensing is enabled, otherwise return False.

        When sensing is enabled, the instrument operates in 4-wire mode, using the SENSE input terminals.
        When sensing is disabled, the instrument operates in 2-wire mode and the SENSE inputs are ignored.
        """
        tok = self._ask_token(cmd="SENS?", tokens=["TWOWIRE", "FOURWIRE"])
        return tok != 0

    @rpc_method
    def set_sensing_enabled(self, enable: bool) -> None:
        """Enable or disable remote sensing mode."""
        self._set_command(f"SENS {int(enable)}")

    @rpc_method
    def get_interlock_status(self) -> bool:
        """Return True if the safety interlock is closed, False if the interlock is open.

        The interlock needs to be closed to operate the instrument in the 100V range.
        """
        tok = self._ask_token(cmd="ILOC?", tokens=["OPEN", "CLOSED"])
        return tok != 0

    @rpc_method
    def get_overloaded(self) -> bool:
        """Return True if the instrument is in overload condition, otherwise return False."""
        tok = self._ask_token(cmd="OVLD?", tokens=["OKAY", "OVLD"])
        return tok != 0



"""Unit tests for Toptica DLC instrument driver."""

import unittest
from unittest.mock import Mock, patch, ANY

from qmi.instruments.toptica import Toptica_DlcPro


class TestTopticaDlc(unittest.TestCase):
    """Unit tests for DLC."""

    def setUp(self):
        self.transport = Mock(name="transport")

        patcher = patch("qmi.instruments.toptica.dlc.create_transport")
        patched_create_transport = patcher.start()
        patched_create_transport.return_value = self.transport
        self.addCleanup(patcher.stop)

        transport_string = "some_transport"
        self.dlc = Toptica_DlcPro(Mock(name="context"), "dlc_under_test", transport_string)
        patched_create_transport.assert_called_with(transport_string)

    def test_open_close(self):
        """Test initialization, open, close."""
        self.dlc.open()
        self.dlc.close()

        self.transport.open.assert_called_once()
        self.transport.read_until.assert_called_once_with(
            message_terminator=b"DeCoF Command Line\r\n> ", timeout=ANY)
        self.transport.close.assert_called_once()

    def test_open_fail(self):
        """Test failed open."""
        self.transport.read_until.side_effect = ValueError

        with self.assertRaises(Exception):
            self.dlc.open()

        self.transport.open.assert_called_once()
        self.transport.close.assert_called_once()

    def test_get_parameter_float(self):
        """Test parameter getter."""
        parameter_name = "my_param"
        parameter_value = 123.456
        self.transport.read_until.side_effect = [
            f"{parameter_value}\n".encode(),
            b"> \n"  # prompt
        ]

        retval = self.dlc.get_parameter(parameter_name)
        self.transport.write.assert_any_call(f"(param-ref '{parameter_name})\n".encode())
        self.assertEqual(retval, parameter_value)

    def test_get_parameter_int(self):
        """Test parameter getter."""
        parameter_name = "my_param"
        parameter_value = 123
        self.transport.read_until.side_effect = [
            f"{parameter_value}\n".encode(),
            b"> \n"  # prompt
        ]

        retval = self.dlc.get_parameter(parameter_name)
        self.transport.write.assert_any_call(f"(param-ref '{parameter_name})\n".encode())
        self.assertEqual(retval, parameter_value)

    def test_get_parameter_string(self):
        """Test parameter getter."""
        parameter_name = "my_param"
        parameter_value = "my_value"
        self.transport.read_until.side_effect = [
            f"\"{parameter_value}\"\n".encode(),
            b"> \n"  # prompt
        ]

        retval = self.dlc.get_parameter(parameter_name)
        self.transport.write.assert_any_call(f"(param-ref '{parameter_name})\n".encode())
        self.assertEqual(retval, parameter_value)

    def test_get_parameter_bool_true(self):
        """Test parameter getter."""
        parameter_name = "my_param"
        parameter_value = True
        self.transport.read_until.side_effect = [
            f"#t\n".encode(),
            b"> \n"  # prompt
        ]

        retval = self.dlc.get_parameter(parameter_name)
        self.transport.write.assert_any_call(f"(param-ref '{parameter_name})\n".encode())
        self.assertEqual(retval, parameter_value)

    def test_get_parameter_bool_false(self):
        """Test parameter getter."""
        parameter_name = "my_param"
        parameter_value = False
        self.transport.read_until.side_effect = [
            f"#f\n".encode(),
            b"> \n"  # prompt
        ]

        retval = self.dlc.get_parameter(parameter_name)
        self.transport.write.assert_any_call(f"(param-ref '{parameter_name})\n".encode())
        self.assertEqual(retval, parameter_value)

    def test_get_parameter_tuple(self):
        """Test parameter getter."""
        parameter_name = "my_param"
        parameter_value = (12, 34.56, "hello")
        self.transport.read_until.side_effect = [
            f"{parameter_value}\n".replace('\'', '"').replace(',', ' ').encode(),  # strings must have double quotes!
            b"> \n"  # prompt
        ]

        retval = self.dlc.get_parameter(parameter_name)
        self.transport.write.assert_any_call(f"(param-ref '{parameter_name})\n".encode())
        self.assertEqual(retval, parameter_value)

    def test_get_multiple_parameter(self):
        """Test parameter getter."""
        parameter_names = ["my_param1", "my_param2", "my_param3"]
        parameter_values = (123, 456.789, "hello")
        self.transport.read_until.side_effect = [
            f"{parameter_values}\n".replace('\'', '"').replace(',', ' ').encode(),  # strings must have double quotes!
            b"> \n"  # prompt
        ]

        retval = self.dlc.get_multiple_parameters(parameter_names)
        self.transport.write.assert_any_call(
            "`(,(param-ref '{p[0]}),(param-ref '{p[1]}),(param-ref '{p[2]}))\n".format(p=parameter_names).encode()
        )
        self.assertEqual(retval, parameter_values)

    def test_get_parameter_error(self):
        """Test parameter getter."""
        self.transport.read_until.side_effect = [
            b"Error: this is an error\n",
            b"> \n"  # prompt
        ]
        with self.assertRaises(ValueError):
            self.dlc.get_parameter("my_param")

    def test_set_parameter_float(self):
        """Test parameter setter."""
        parameter_name = "my_param"
        parameter_value = 123.456
        self.transport.read_until.side_effect = [
            b"0\n",  # success
            b"> \n"  # prompt
        ]

        self.dlc.set_parameter(parameter_name, parameter_value)
        self.transport.write.assert_any_call(f"(param-set! '{parameter_name} {parameter_value})\n".encode())

    def test_set_parameter_int(self):
        """Test parameter setter."""
        parameter_name = "my_param"
        parameter_value = 123
        self.transport.read_until.side_effect = [
            b"0\n",  # success
            b"> \n"  # prompt
        ]

        self.dlc.set_parameter(parameter_name, parameter_value)
        self.transport.write.assert_any_call(f"(param-set! '{parameter_name} {parameter_value})\n".encode())

    def test_set_parameter_string(self):
        """Test parameter setter."""
        parameter_name = "my_param"
        parameter_value = "my_value"
        self.transport.read_until.side_effect = [
            b"0\n",  # success
            b"> \n"  # prompt
        ]

        self.dlc.set_parameter(parameter_name, parameter_value)
        self.transport.write.assert_any_call(f"(param-set! '{parameter_name} {parameter_value})\n".encode())

    def test_set_parameter_bool_true(self):
        """Test parameter setter."""
        parameter_name = "my_param"
        parameter_value = True
        self.transport.read_until.side_effect = [
            b"0\n",  # success
            b"> \n"  # prompt
        ]

        self.dlc.set_parameter(parameter_name, parameter_value)
        self.transport.write.assert_any_call(f"(param-set! '{parameter_name} #t)\n".encode())

    def test_set_parameter_bool_false(self):
        """Test parameter setter."""
        parameter_name = "my_param"
        parameter_value = False
        self.transport.read_until.side_effect = [
            b"0\n",  # success
            b"> \n"  # prompt
        ]

        self.dlc.set_parameter(parameter_name, parameter_value)
        self.transport.write.assert_any_call(f"(param-set! '{parameter_name} #f)\n".encode())

    def test_set_multiple_parameter(self):
        """Test parameter setter."""
        parameter_names = ["my_param1", "my_param2", "my_param3"]
        parameter_values = (123, 456.789, "hello")
        self.transport.read_until.side_effect = [
            b"0\n",
            b"> \n"  # prompt
        ]

        self.dlc.set_multiple_parameters(zip(parameter_names, parameter_values))
        self.transport.write.assert_any_call(
            "(+ (param-set! '{p[0]} {v[0]})(param-set! '{p[1]} {v[1]})(param-set! '{p[2]} {v[2]}))\n".format(
                p=parameter_names, v=parameter_values
            ).encode()
        )

    def test_set_parameter_error(self):
        """Test parameter setter."""
        self.transport.read_until.side_effect = [
            b"Error: this is an error\n",
            b"> \n"  # prompt
        ]
        with self.assertRaises(ValueError):
            self.dlc.set_parameter("my_param", 123.456)



"""Error codes returned by PI GCS-enabled instruments."""

# Source: SM158E Software Manual PI GCS2 Driver Library for NI LabVIEW, release: 7.0.1, date: 2021-03-03

# Controller errors
GCS_CONTROLLER_ERROR_CODES = {
    0: "PI_CNTR_NO_ERROR",
    1: "PI_CNTR_PARAM_SYNTAX",
    2: "PI_CNTR_UNKNOWN_COMMAND",
    3: "PI_CNTR_COMMAND_TOO_LONG",
    4: "PI_CNTR_SCAN_ERROR",
    5: "PI_CNTR_MOVE_WITHOUT_REF_OR_NO_SERVO",
    6: "PI_CNTR_INVALID_SGA_PARAM",
    7: "PI_CNTR_POS_OUT_OF_LIMITS",
    8: "PI_CNTR_VEL_OUT_OF_LIMITS",
    9: "PI_CNTR_SET_PIVOT_NOT_POSSIBLE",
    10: "PI_CNTR_STOP",
    11: "PI_CNTR_SST_OR_SCAN_RANGE",
    12: "PI_CNTR_INVALID_SCAN_AXES",
    13: "PI_CNTR_INVALID_NAV_PARAM",
    14: "PI_CNTR_INVALID_ANALOG_INPUT",
    15: "PI_CNTR_INVALID_AXIS_IDENTIFIER",
    16: "PI_CNTR_INVALID_STAGE_NAME",
    17: "PI_CNTR_PARAM_OUT_OF_RANGE",
    18: "PI_CNTR_INVALID_MACRO_NAME",
    19: "PI_CNTR_MACRO_RECORD",
    20: "PI_CNTR_MACRO_NOT_FOUND",
    21: "PI_CNTR_AXIS_HAS_NO_BRAKE",
    22: "PI_CNTR_DOUBLE_AXIS",
    23: "PI_CNTR_ILLEGAL_AXIS",
    24: "PI_CNTR_PARAM_NR",
    25: "PI_CNTR_INVALID_REAL_NR",
    26: "PI_CNTR_MISSING_PARAM",
    27: "PI_CNTR_SOFT_LIMIT_OUT_OF_RANGE",
    28: "PI_CNTR_NO_MANUAL_PAD",
    29: "PI_CNTR_NO_JUMP",
    30: "PI_CNTR_INVALID_JUMP",
    31: "PI_CNTR_AXIS_HAS_NO_REFERENCE",
    32: "PI_CNTR_STAGE_HAS_NO_LIM_SWITCH",
    33: "PI_CNTR_NO_RELAY_CARD",
    34: "PI_CNTR_CMD_NOT_ALLOWED_FOR_STAGE",
    35: "PI_CNTR_NO_DIGITAL_INPUT",
    36: "PI_CNTR_NO_DIGITAL_OUTPUT",
    37: "PI_CNTR_NO_MCM",
    38: "PI_CNTR_INVALID_MCM",
    39: "PI_CNTR_INVALID_CNTR_NUMBER",
    40: "PI_CNTR_NO_JOYSTICK_CONNECTED",
    41: "PI_CNTR_INVALID_EGE_AXIS",
    42: "PI_CNTR_SLAVE_POSITION_OUT_OF_RANGE",
    43: "PI_CNTR_COMMAND_EGE_SLAVE",
    44: "PI_CNTR_JOYSTICK_CALIBRATION_FAILED",
    45: "PI_CNTR_REFERENCING_FAILED",
    46: "PI_CNTR_OPM_MISSING",
    47: "PI_CNTR_OPM_NOT_INITIALIZED",
    48: "PI_CNTR_OPM_COM_ERROR",
    49: "PI_CNTR_MOVE_TO_LIMIT_SWITCH_FAILED",
    50: "PI_CNTR_REF_WITH_REF_DISABLED",
    51: "PI_CNTR_AXIS_UNDER_JOYSTICK_CONTROL",
    52: "PI_CNTR_COMMUNICATION_ERROR",
    53: "PI_CNTR_DYNAMIC_MOVE_IN_PROGRESS",
    54: "PI_CNTR_UNKNOWN_PARAMETER",
    55: "PI_CNTR_NO_REP_RECORDED",
    56: "PI_CNTR_INVALID_PASSWORD",
    57: "PI_CNTR_INVALID_RECORDER_CHAN",
    58: "PI_CNTR_INVALID_RECORDER_SRC_OPT",
    59: "PI_CNTR_INVALID_RECORDER_SRC_CHAN",
    60: "PI_CNTR_PARAM_PROTECTION",
    61: "PI_CNTR_AUTOZERO_RUNNING",
    62: "PI_CNTR_NO_LINEAR_AXIS",
    63: "PI_CNTR_INIT_RUNNING",
    64: "PI_CNTR_READ_ONLY_PARAMETER",
    65: "PI_CNTR_PAM_NO_FOUND",
    66: "PI_CNTR_VOL_OUT_OF_LIMITS",
    67: "PI_CNTR_WAVE_TOO_LARGE",
    68: "PI_CNTR_NOT_ENOUGH_DDL_MEMORY",
    69: "PI_CNTR_DDL_TIME_DELAY_TOO_LARGE",
    70: "PI_CNTR_DIFFERENT_ARRAY_LENGTH",
    71: "PI_CNTR_GEN_SINGLE_MODE_RESTART",
    72: "PI_CNTR_ANALOG_TARGET_ACTIVE",
    73: "PI_CNTR_WAVE_GENERATOR_ACTIVE",
    74: "PI_CNTR_AUTOZERO_DISABLED",
    75: "PI_CNTR_NO_WAVE_SELECTED",
    76: "PI_CNTR_IF_BUFFER_OVERRUN",
    77: "PI_CNTR_NOT_ENOUGH_RECORDED_DATA",
    78: "PI_CNTR_TABLE_DEACTIVATED",
    79: "PI_CNTR_OPENLOOP_VALUE_SE_WHEN_SERVO_ON",
    80: "PI_CNTR_RAM_ERROR",
    81: "PI_CNTR_MACRO_UNKNOWN_COMMAND",
    82: "PI_CNTR_MACRO_PC_ERROR",
    83: "PI_CNTR_JOYSTICK_ACTIVE",
    84: "PI_CNTR_MOTOR_IS_OFF",
    85: "PI_CNTR_ONLY_IN_MACRO",
    86: "PI_CNTR_JOYSTCIK_UNKNOWN_AXIS",
    87: "PI_CNTR_JOYSTICK_UNKNOWN_ID",
    88: "PI_CNTR_REF_MODE_IS_ON",
    89: "PI_CNTR_NOT_ALLOWED_IN_CURRENT_MOTION_MODE",
    90: "PI_CNTR_DIO_AND_TRACING_NOT_POSSIBLE",
    91: "PI_CNTR_COLLISION",
    92: "PI_CNTR_SLAVE_NOT_FAST_ENOUGH",
    93: "PI_CNTR_COMMAND_NOT_ALLOWED_WHILE_AXIS_IN_MOTION",
    94: "PI_CNTR_OPEN_LOOP_JOYSTICK_ENABLED",
    95: "PI_CNTR_INVALID_SERVO_STATE_FOR_PARAMETER",
    96: "PI_CNTR_UNKNOWN_STAGE_NAME",
    97: "PI_CNTR_INVALID_VALUE_LENGTH",
    98: "PI_CNTR_AUTOZERO_FAILED",
    99: "PI_CNTR_SENSOR_VOLTAGE_OFF",
    100: "PI_LABVIEW_ERROR",
    200: "PI_CNTR_NO_AXIS",
    201: "PI_CNTR_NO_AXIS_PARAM_FILE",
    202: "PI_CNTR_INVALID_AXIS_PARAM_FILE",
    203: "PI_CNTR_NO_AXIS_PARAM_BACKUP",
    204: "PI_CNTR_RESERVED_204",
    205: "PI_CNTR_SMO_WITH_SERVO_ON",
    206: "PI_CNTR_UUDECODE_INCOMPLETE_HEADER",
    207: "PI_CNTR_UUDECODE_NOTHING_TO_DECODE",
    208: "PI_CNTR_UUDECODE_ILLEGAL_FORMAT",
    209: "PI_CNTR_CRC32_ERROR",
    210: "PI_CNTR_ILLEGAL_FILENAME",
    211: "PI_CNTR_FILE_NOT_FOUND",
    212: "PI_CNTR_FILE_WRITE_ERROR",
    213: "PI_CNTR_DTR_HINDERS_VELOCITY_CHANGE",
    214: "PI_CNTR_POSITION_UNKNOWN",
    215: "PI_CNTR_CONN_POSSIBLY_BROKEN",
    216: "PI_CNTR_ON_LIMIT_SWITCH",
    217: "PI_CNTR_UNEXPECTED_STRUT_STOP",
    218: "PI_CNTR_POSITION_BASED_ON_ESTIMATION",
    219: "PI_CNTR_POSITION_BASED_ON_INTERPOLATION",
    220: "PI_CNTR_INTERPOLATION_FIFO_UNDERRUN",
    221: "PI_CNTR_INTERPOLATION_FIFO_OVERFLOW",
    230: "PI_CNTR_INVALID_HANDLE",
    231: "PI_CNTR_NO_BIOS_FOUND",
    232: "PI_CNTR_SAVE_SYS_CFG_FAILED",
    233: "PI_CNTR_LOAD_SYS_CFG_FAILED",
    301: "PI_CNTR_SEND_BUFFER_OVERFLOW",
    302: "PI_CNTR_VOLTAGE_OUT_OF_LIMITS",
    303: "PI_CNTR_OPEN_LOOP_MOTION_SET_WHEN_SERVO_ON",
    304: "PI_CNTR_RECEIVING_BUFFER_OVERFLOW",
    305: "PI_CNTR_EEPROM_ERROR",
    306: "PI_CNTR_I2C_ERROR",
    307: "PI_CNTR_RECEIVING_TIMEOUT",
    308: "PI_CNTR_TIMEOUT",
    309: "PI_CNTR_MACRO_OUT_OF_SPACE",
    310: "PI_CNTR_EUI_OLDVERSION_CFGDATA",
    311: "PI_CNTR_EUI_INVALID_CFGDATA",
    333: "PI_CNTR_HARDWARE_ERROR",
    400: "PI_CNTR_WAV_INDEX_ERROR",
    401: "PI_CNTR_WAV_NOT_DEFINED",
    402: "PI_CNTR_WAV_TYPE_NOT_SUPPORTED",
    403: "PI_CNTR_WAV_LENGTH_EXCEEDS_LIMIT",
    404: "PI_CNTR_WAV_PARAMETER_NR",
    405: "PI_CNTR_WAV_PARAMETER_OUT_OF_LIMIT",
    406: "PI_CNTR_WGO_BIT_NOT_SUPPORTED",
    500: "PI_CNTR_EMERGENCY_STOP_BUTTON_ACTIVATED",
    501: "PI_CNTR_EMERGENCY_STOP_BUTTON_WAS_ACTIVATED",
    502: "PI_CNTR_REDUNDANCY_LIMIT_EXCEEDED",
    503: "PI_CNTR_COLLISION_SWITCH_ACTIVATED",
    504: "PI_CNTR_FOLLOWING_ERROR",
    505: "PI_CNTR_SENSOR_SIGNAL_INVALID",
    506: "PI_CNTR_SERVO_LOOP_UNSTABLE",
    507: "PI_CNTR_LOST_SPI_SLAVE_CONNECTION",
    508: "PI_CNTR_MOVE_ATTEMPT_NOT_PERMITTED",
    509: "PI_CNTR_TRIGGER_EMERGENCY_STOP",
    530: "PI_CNTR_CS_DOES_NOT_EXIST",
    531: "PI_CNTR_PARENT_CS_DOES_NOT_EXIST",
    532: "PI_CNTR_CS_IN_USE",
    533: "PI_CNTR_CS_DEFINITION_IS_CYCLIC",
    536: "PI_CNTR_HEXAPOD_IN_MOTION",
    537: "PI_CNTR_CS_TYPE_CANNOT_BE_ENABLED",
    539: "PI_CNTR_CS_PARENT_IDENTICAL_TO_CHILD",
    540: "PI_CNTR_CS_DEFINITION_INCONSISTENT",
    542: "PI_CNTR_CS_NOT_IN_SAME_CHAIN",
    543: "PI_CNTR_CS_MEMORY_FULL",
    544: "PI_CNTR_SPI_COMMAND_NOT_SUPPORTED",
    545: "PI_CNTR_SOFTLIMITS_INVALID",
    546: "PI_CNTR_CS_WRITE_PROTECTED",
    547: "PI_CNTR_CS_CONTENT_FROM_CONFIG_FILE",
    548: "PI_CNTR_CS_CANNOT_BE_LINKED",
    549: "PI_CNTR_KSB_CS_ROTATION_ONLY",
    551: "PI_CNTR_CS_DATA_CANNOT_BE_QUERIED",
    552: "PI_CNTR_CS_COMBINATION_DOES_NOT_EXIST",
    553: "PI_CNTR_CS_COMBINATION_INVALID",
    554: "PI_CNTR_CS_TYPE_DOES_NOT_EXIST",
    555: "PI_CNTR_UNKNOWN_ERROR",
    556: "PI_CNTR_CS_TYPE_NOT_ENABLED",
    557: "PI_CNTR_CS_NAME_INVALID",
    558: "PI_CNTR_CS_GENERAL_FILE_MISSING",
    559: "PI_CNTR_CS_LEVELING_FILE_MISSING",
    601: "PI_CNTR_NOT_ENOUGH_MEMORY",
    602: "PI_CNTR_HW_VOLTAGE_ERROR",
    603: "PI_CNTR_HW_TEMPERATURE_ERROR",
    604: "PI_CNTR_POSITION_ERROR_TOO_HIGH",
    606: "PI_CNTR_INPUT_OUT_OF_RANGE",
    607: "PI_CNTR_NO_INTEGER",
    608: "PI_CNTR_FAST_ALIGNMENT_PROCESS_IS_NOT_RUNNING",
    609: "PI_CNTR_FAST_ALIGNMENT_PROCESS_IS_NOT_PAUSED",
    650: "PI_CNTR_UNABLE_TO_SET_PARAM_WITH_SPA",
    651: "PI_CNTR_PHASE_FINDING_ERROR",
    652: "PI_CNTR_SENSOR_SETUP_ERROR",
    653: "PI_CNTR_SENSOR_COMM_ERROR",
    654: "PI_CNTR_MOTOR_AMPLIFIER_ERROR",
    655: "PI_CNTR_OVER_CURR_PROTEC_TRIGGERED_BY_I2T",
    656: "PI_CNTR_OVER_CURR_PROTEC_TRIGGERED_BY_AMP_MODULE",
    657: "PI_CNTR_SAFETY_STOP_TRIGGERED",
    658: "PI_SENSOR_OFF",
    700: "PI_CNTR_COMMAND_NOT_ALLOWED_IN_EXTERNAL_MODE",
    710: "PI_CNTR_EXTERNAL_MODE_ERROR",
    715: "PI_CNTR_INVALID_MODE_OF_OPERATION",
    716: "PI_CNTR_FIRMWARE_STOPPED_BY_CMD",
    717: "PI_CNTR_EXTERNAL_MODE_DRIVER_MISSING",
    718: "PI_CNTR_CONFIGURATION_FAILURE_EXTERNAL_MODE",
    719: "PI_CNTR_EXTERNAL_MODE_CYCLETIME_INVALID",
    720: "PI_CNTR_BRAKE_ACTIVATED",
    731: "PI_CNTR_SURFACEDETECTION_RUNNING",
    732: "PI_CNTR_SURFACEDETECTION_FAILED",
    733: "PI_CNTR_FIELDBUS_IS_ACTIVE",
    1000: "PI_CNTR_TOO_MANY_NESTED_MACROS",
    1001: "PI_CNTR_MACRO_ALREADY_DEFINED",
    1002: "PI_CNTR_NO_MACRO_RECORDING",
    1003: "PI_CNTR_INVALID_MAC_PARAM",
    1004: "PI_CNTR_MACRO_DELETE_ERROR",
    1005: "PI_CNTR_CONTROLLER_BUSY",
    1006: "PI_CNTR_INVALID_IDENTIFIER",
    1007: "PI_CNTR_UNKNOWN_VARIABLE_OR_ARGUMENT",
    1008: "PI_CNTR_RUNNING_MACRO",
    1009: "PI_CNTR_MACRO_INVALID_OPERATOR",
    1010: "PI_CNTR_MACRO_NO_ANSWER",
    1011: "PI_CMD_NOT_VALID_IN_MACRO_MODE",
    1012: "PI_CNTR_ERROR_IN_MACRO",
    1024: "PI_CNTR_MOTION_ERROR",
    1025: "PI_CNTR_MAX_MOTOR_OUTPUT_REACHED",
    1063: "PI_CNTR_EXT_PROFILE_UNALLOWED_CMD",
    1064: "PI_CNTR_EXT_PROFILE_EXPECTING_MOTION_ERROR",
    1065: "PI_CNTR_PROFILE_ACTIVE",
    1066: "PI_CNTR_PROFILE_INDEX_OUT_OF_RANGE",
    1071: "PI_CNTR_PROFILE_OUT_OF_MEMORY",
    1072: "PI_CNTR_PROFILE_WRONG_CLUSTER",
    1073: "PI_CNTR_PROFILE_UNKNOWN_CLUSTER_IDENTIFIER",
    1090: "PI_CNTR_TOO_MANY_TCP_CONNECTIONS_OPEN",
    2000: "PI_CNTR_ALREADY_HAS_SERIAL_NUMBER",
    4000: "PI_CNTR_SECTOR_ERASE_FAILED",
    4001: "PI_CNTR_FLASH_PROGRAM_FAILED",
    4002: "PI_CNTR_FLASH_READ_FAILED",
    4003: "PI_CNTR_HW_MATCHCODE_ERROR",
    4004: "PI_CNTR_FW_MATCHCODE_ERROR",
    4005: "PI_CNTR_HW_VERSION_ERROR",
    4006: "PI_CNTR_FW_VERSION_ERROR",
    4007: "PI_CNTR_FW_UPDATE_ERROR",
    4008: "PI_CNTR_FW_CRC_PAR_ERROR",
    4009: "PI_CNTR_FW_CRC_FW_ERROR",
    5000: "PI_CNTR_INVALID_PCC_SCAN_DATA",
    5001: "PI_CNTR_PCC_SCAN_RUNNING",
    5002: "PI_CNTR_INVALID_PCC_AXIS",
    5003: "PI_CNTR_PCC_SCAN_OUT_OF_RANGE",
    5004: "PI_CNTR_PCC_TYPE_NOT_EXISTING",
    5005: "PI_CNTR_PCC_PAM_ERROR",
    5006: "PI_CNTR_PCC_TABLE_ARRAY_TOO_LARGE",
    5100: "PI_CNTR_NEXLINE_ERROR",
    5101: "PI_CNTR_CHANNEL_ALREADY_USED",
    5102: "PI_CNTR_NEXLINE_TABLE_TOO_SMALL",
    5103: "PI_CNTR_RNP_WITH_SERVO_ON",
    5104: "PI_CNTR_RNP_NEEDED",
    5200: "PI_CNTR_AXIS_NOT_CONFIGURED",
    5300: "PI_CNTR_FREQU_ANALYSIS_FAILED",
    5301: "PI_CNTR_FREQU_ANALYSIS_RUNNING",
    6000: "PI_CNTR_SENSOR_ABS_INVALID_VALUE",
    6001: "PI_CNTR_SENSOR_ABS_WRITE_ERROR",
    6002: "PI_CNTR_SENSOR_ABS_READ_ERROR",
    6003: "PI_CNTR_SENSOR_ABS_CRC_ERROR",
    6004: "PI_CNTR_SENSOR_ABS_ERROR",
    6005: "PI_CNTR_SENSOR_ABS_OVERFLOW"
}

# Interface errors
GCS_INTERFACE_ERROR_CODES = {
    0: "COM_NO_ERROR",
    -1: "COM_ERROR",
    -2: "SEND_ERROR",
    -3: "REC_ERROR",
    -4: "NOT_CONNECTED_ERROR",
    -5: "COM_BUFFER_OVERFLOW",
    -6: "CONNECTION_FAILED",
    -7: "COM_TIMEOUT",
    -8: "COM_MULTILINE_RESPONSE",
    -9: "COM_INVALID_ID",
    -10: "COM_NOTIFY_EVENT_ERROR",
    -11: "COM_NOT_IMPLEMENTED",
    -12: "COM_ECHO_ERROR",
    -13: "COM_GPIB_EDVR",
    -14: "COM_GPIB_ECIC",
    -15: "COM_GPIB_ENOL",
    -16: "COM_GPIB_EADR",
    -17: "COM_GPIB_EARG",
    -18: "COM_GPIB_ESAC",
    -19: "COM_GPIB_EABO",
    -20: "COM_GPIB_ENEB",
    -21: "COM_GPIB_EDMA",
    -22: "COM_GPIB_EOIP",
    -23: "COM_GPIB_ECAP",
    -24: "COM_GPIB_EFSO",
    -25: "COM_GPIB_EBUS",
    -26: "COM_GPIB_ESTB",
    -27: "COM_GPIB_ESRQ",
    -28: "COM_GPIB_ETAB",
    -29: "COM_GPIB_ECLK",
    -30: "COM_RS_INVALID_DATA_BITS",
    -31: "COM_ERROR_RS_SETTINGS",
    -32: "COM_INTERNAL_RESOURCES_ERROR",
    -33: "COM_DLL_FUNC_ERROR",
    -34: "COM_FTDIUSB_INVALID_HANDLE",
    -35: "COM_FTDIUSB_DEVICE_NOT_FOUND",
    -36: "COM_FTDIUSB_DEVICE_NOT_OPENED",
    -37: "COM_FTDIUSB_IO_ERROR",
    -38: "COM_FTDIUSB_INSUFFICIENT_RESOURCES",
    -39: "COM_FTDIUSB_INVALID_PARAMETER",
    -40: "COM_FTDIUSB_INVALID_BAUD_RATE",
    -41: "COM_FTDIUSB_DEVICE_NOT_OPENED_FOR_ERASE",
    -42: "COM_FTDIUSB_DEVICE_NOT_OPENED_FOR_WRITE",
    -43: "COM_FTDIUSB_FAILED_TO_WRITE_DEVICE",
    -44: "COM_FTDIUSB_EEPROM_READ_FAILED",
    -45: "COM_FTDIUSB_EEPROM_WRITE_FAILED",
    -46: "COM_FTDIUSB_EEPROM_ERASE_FAILED",
    -47: "COM_FTDIUSB_EEPROM_NOT_PRESENT",
    -48: "COM_FTDIUSB_EEPROM_NOT_PROGRAMMED",
    -49: "COM_FTDIUSB_INVALID_ARGS",
    -50: "COM_FTDIUSB_NOT_SUPPORTED",
    -51: "COM_FTDIUSB_OTHER_ERROR",
    -52: "COM_PORT_ALREADY_OPEN",
    -53: "COM_PORT_CHECKSUM_ERROR",
    -54: "COM_SOCKET_NOT_READY",
    -55: "COM_SOCKET_PORT_IN_USE",
    -56: "COM_SOCKET_NOT_CONNECTED",
    -57: "COM_SOCKET_TERMINATED",
    -58: "COM_SOCKET_NO_RESPONSE",
    -59: "COM_SOCKET_INTERRUPTED",
    -60: "COM_PCI_INVALID_ID",
    -61: "COM_PCI_ACCESS_DENIED",
    -62: "COM_SOCKET_HOST_NOT_FOUND",
    -63: "COM_DEVICE_CONNECTED"
}

# DLL errors
GCS_DLL_ERROR_CODES = {
    -1001: "PI_UNKNOWN_AXIS_IDENTIFIER",
    -1002: "PI_NR_NAV_OUT_OF_RANGE",
    -1003: "PI_INVALID_SGA",
    -1004: "PI_UNEXPECTED_RESPONSE",
    -1005: "PI_NO_MANUAL_PAD",
    -1006: "PI_INVALID_MANUAL_PAD_KNOB",
    -1007: "PI_INVALID_MANUAL_PAD_AXIS",
    -1008: "PI_CONTROLLER_BUSY",
    -1009: "PI_THREAD_ERROR",
    -1010: "PI_IN_MACRO_MODE",
    -1011: "PI_NOT_IN_MACRO_MODE",
    -1012: "PI_MACRO_FILE_ERROR",
    -1013: "PI_NO_MACRO_OR_EMPTY",
    -1014: "PI_MACRO_EDITOR_ERROR",
    -1015: "PI_INVALID_ARGUMENT",
    -1016: "PI_AXIS_ALREADY_EXISTS",
    -1017: "PI_INVALID_AXIS_IDENTIFIER",
    -1018: "PI_COM_ARRAY_ERROR",
    -1019: "PI_COM_ARRAY_RANGE_ERROR",
    -1020: "PI_INVALID_SPA_CMD_ID",
    -1021: "PI_NR_AVG_OUT_OF_RANGE",
    -1022: "PI_WAV_SAMPLES_OUT_OF_RANGE",
    -1023: "PI_WAV_FAILED",
    -1024: "PI_MOTION_ERROR",
    -1025: "PI_RUNNING_MACRO",
    -1026: "PI_PZT_CONFIG_FAILED",
    -1027: "PI_PZT_CONFIG_INVALID_PARAMS",
    -1028: "PI_UNKNOWN_CHANNEL_IDENTIFIER",
    -1029: "PI_WAVE_PARAM_FILE_ERROR",
    -1030: "PI_UNKNOWN_WAVE_SET",
    -1031: "PI_WAVE_EDITOR_FUNC_NOT_LOADED",
    -1032: "PI_USER_CANCELLED",
    -1033: "PI_C844_ERROR",
    -1034: "PI_DLL_NOT_LOADED",
    -1035: "PI_PARAMETER_FILE_PROTECTED",
    -1036: "PI_NO_PARAMETER_FILE_OPENED",
    -1037: "PI_STAGE_DOES_NOT_EXIST",
    -1038: "PI_PARAMETER_FILE_ALREADY_OPENED",
    -1039: "PI_PARAMETER_FILE_OPEN_ERROR",
    -1040: "PI_INVALID_CONTROLLER_VERSION",
    -1041: "PI_PARAM_SET_ERROR",
    -1042: "PI_NUMBER_OF_POSSIBLE_WAVES_EXCEEDED",
    -1043: "PI_NUMBER_OF_POSSIBLE_GENERATORS_EXCEEDED",
    -1044: "PI_NO_WAVE_FOR_AXIS_DEFINED",
    -1045: "PI_CANT_STOP_OR_START_WAV",
    -1046: "PI_REFERENCE_ERROR",
    -1047: "PI_REQUIRED_WAVE_NOT_FOUND",
    -1048: "PI_INVALID_SPP_CMD_ID",
    -1049: "PI_STAGE_NAME_ISNT_UNIQUE",
    -1050: "PI_FILE_TRANSFER_BEGIN_MISSING",
    -1051: "PI_FILE_TRANSFER_ERROR_TEMP_FILE",
    -1052: "PI_FILE_TRANSFER_CRC_ERROR",
    -1053: "PI_COULDNT_FIND_PISTAGES_DAT",
    -1054: "PI_NO_WAVE_RUNNING",
    -1055: "PI_INVALID_PASSWORD",
    -1056: "PI_OPM_COM_ERROR",
    -1057: "PI_WAVE_EDITOR_WRONG_PARAMNUM",
    -1058: "PI_WAVE_EDITOR_FREQUENCY_OUT_OF_RANGE",
    -1059: "PI_WAVE_EDITOR_WRONG_IP_VALUE",
    -1060: "PI_WAVE_EDITOR_WRONG_DP_VALUE",
    -1061: "PI_WAVE_EDITOR_WRONG_ITEM_VALUE",
    -1062: "PI_WAVE_EDITOR_MISSING_GRAPH_COMPONENT",
    -1063: "PI_EXT_PROFILE_UNALLOWED_CMD",
    -1064: "PI_EXT_PROFILE_EXPECTING_MOTION_ERROR",
    -1065: "PI_EXT_PROFILE_ACTIVE",
    -1066: "PI_EXT_PROFILE_INDEX_OUT_OF_RANGE",
    -1067: "PI_PROFILE_GENERATOR_NO_PROFILE",
    -1068: "PI_PROFILE_GENERATOR_OUT_OF_LIMITS",
    -1069: "PI_PROFILE_GENERATOR_UNKNOWN_PARAMETER",
    -1070: "PI_PROFILE_GENERATOR_PAR_OUT_OF_RANGE",
    -1071: "PI_EXT_PROFILE_OUT_OF_MEMORY",
    -1072: "PI_EXT_PROFILE_WRONG_CLUSTER",
    -1073: "PI_UNKNOWN_CLUSTER_IDENTIFIER",
    -1074: "PI_INVALID_DEVICE_DRIVER_VERSION",
    -1075: "PI_INVALID_LIBRARY_VERSION",
    -1076: "PI_INTERFACE_LOCKED",
    -1077: "PI_PARAM_DAT_FILE_INVALID_VERSION",
    -1078: "PI_CANNOT_WRITE_TO_PARAM_DAT_FILE",
    -1079: "PI_CANNOT_CREATE_PARAM_DAT_FILE",
    -1080: "PI_PARAM_DAT_FILE_INVALID_REVISION",
    -1081: "PI_USERSTAGES_DAT_FILE_INVALID_REVISION",
    -1082: "PI_SOFTWARE_TIMEOUT",
    -1083: "PI_WRONG_DATA_TYPE",
    -1084: "PI_DIFFERENT_ARRAY_SIZES",
    -1085: "PI_PARAM_NOT_FOUND_IN_PARAM_DAT_FILE",
    -1086: "PI_MACRO_RECORDING_NOT_ALLOWED_IN_THIS_MODE",
    -1087: "PI_USER_CANCELLED_COMMAND",
    -1088: "PI_TOO_FEW_GCS_DATA",
    -1089: "PI_TOO_MANY_GCS_DATA",
    -1090: "PI_GCS_DATA_READ_ERROR",
    -1091: "PI_WRONG_NUMBER_OF_INPUT_ARGUMENTS",
    -1092: "PI_FAILED_TO_CHANGE_CCL_LEVEL",
    -1093: "PI_FAILED_TO_SWITCH_OFF_SERVO",
    -1094: "PI_FAILED_TO_SET_SINGLE_PARAMETER_WHILE_PERFORMING_CST",
    -1095: "PI_ERROR_CONTROLLER_REBOOT",
    -1096: "PI_ERROR_AT_QHPA",
    -1097: "PI_QHPA_NONCOMPLIANT_WITH_GCS",
    -1098: "PI_FAILED_TO_READ_QSPA",
    -1099: "PI_PAM_FILE_WRONG_VERSION",
    -1100: "PI_PAM_FILE_INVALID_FORMAT",
    -1101: "PI_INCOMPLETE_INFORMATION",
    -1102: "PI_NO_VALUE_AVAILABLE",
    -1103: "PI_NO_PAM_FILE_OPEN",
    -1104: "PI_INVALID_VALUE",
    -1105: "PI_UNKNOWN_PARAMETER",
    -1106: "PI_RESPONSE_TO_QSEP_FAILED",
    -1107: "PI_RESPONSE_TO_QSPA_FAILED",
    -1108: "PI_ERROR_IN_CST_VALIDATION",
    -1109: "PI_ERROR_PAM_FILE_HAS_DUPLICATE_ENTRY_WITH_DIFFERENT_VALUES",
    -1110: "PI_ERROR_FILE_NO_SIGNATURE",
    -1111: "PI_ERROR_FILE_INVALID_SIGNATURE",
    -1112: "PI_ERROR_CANNOT_DETERMINE_ACTUAL_END_OF_TRAVEL_WHILE_PLATFORM_IS_MOVING",
    -10000: "PI_PARAMETER_DB_INVALID_STAGE_TYPE_FORMAT",
    -10001: "PI_PARAMETER_DB_SYSTEM_NOT_AVAILABLE",
    -10002: "PI_PARAMETER_DB_FAILED_TO_ESTABLISH_CONNECTION",
    -10003: "PI_PARAMETER_DB_COMMUNICATION_ERROR",
    -10004: "PI_PARAMETER_DB_ERROR_WHILE_QUERYING_PARAMETERS",
    -10005: "PI_PARAMETER_DB_SYSTEM_ALREADY_EXISTS",
    -10006: "PI_PARAMETER_DB_QHPA_CONTAINS_UNKNOWN_PAM_IDS",
    -10007: "PI_PARAMETER_DB_AND_QHPA_ARE_INCONSISTENT",
    -10008: "PI_PARAMETER_DB_SYSTEM_COULD_NOT_BE_ADDED",
    -10009: "PI_PARAMETER_DB_SYSTEM_COULD_NOT_BE_REMOVED",
    -10010: "PI_PARAMETER_DB_CONTROLLER_DB_PARAMETERS_MISMATCH",
    -10011: "PI_PARAMETER_DB_DATABASE_IS_OUTDATED",
    -10012: "PI_PARAMETER_DB_AND_HPA_MISMATCH_STRICT",
    -10013: "PI_PARAMETER_DB_AND_HPA_MISMATCH_LOOSE",
    -10014: "PI_PARAMETER_DB_FAILED_TO_SET_PARAMETERS_CORRECTLY",
    -10015: "PI_PARAMETER_DB_MISSING_PARAMETER_DEFINITIONS_IN_DATABASE"
}



"""
High Finesse, Wavelength meter.

The qmi.instruments.high_finesse package provides support for:
  - HighFinesse WLM Controller model WS-6 (tested).
  - HighFinesse WLM Controller model WS-5, WS-7, WS-8 (untested).
"""
from qmi.instruments.high_finesse.wlm import HighFinesse_Wlm



import unittest
from unittest.mock import MagicMock, patch
import logging
from typing import cast

import numpy as np

import qmi
from qmi.core.context import QMI_Context
from qmi.core.exceptions import (
    QMI_InstrumentException,
    QMI_UsageException,
    QMI_TimeoutException,
)
from qmi.instruments.ptgrey import Flir_Blackfly_Aravis
import qmi.instruments.ptgrey.blackfly_aravis as ba
import tests.instruments.ptgrey.aravis_stub as Aravis


class Region:
    # The returned items from Aravis.Camera.get_region are pointers,
    # not using a specific type. These pointers refer to non-existing
    # in-memory camera, thus using a stubbed Region class.
    x = None
    y = None
    width = None
    height = None


class TestPtGreyBlackFly(unittest.TestCase):
    def _create_instr_obj(self):
        """Create a PtGeyBlackFlyAravis object using an serial_number."""
        instr: Flir_Blackfly_Aravis = Flir_Blackfly_Aravis(
            QMI_Context("context"),
            "name",
            "SerialNumber",
            None,
        )
        return instr

    def _create_instr_ser(self):
        """Create a PtGeyBlackFlyAravis instrument using a serial_number."""
        instr: Flir_Blackfly_Aravis = qmi.make_instrument(
            "name",
            Flir_Blackfly_Aravis,
            "SerialNumber",
            None,
        )
        instr = cast(Flir_Blackfly_Aravis, instr)
        return instr

    def _create_instr_ip(self):
        """Create a PtGeyBlackFlyAravis instrument using an ip_address."""
        instr: Flir_Blackfly_Aravis = qmi.make_instrument(
            "name",
            Flir_Blackfly_Aravis,
            None,
            "10.10.10.10",
        )
        instr = cast(Flir_Blackfly_Aravis, instr)
        return instr

    def _get_feature(self, feat):
        """Retrieves a MagicMock specced to the expected type that
        is mapped to a feature name."""
        return self.feature_map[feat]

    def setUp(self):
        logging.getLogger("qmi.instruments.ptgrey.blackfly_aravis").setLevel(logging.CRITICAL)
        qmi.start("TestContext")
        ba.Aravis = self.aravis = MagicMock(spec=Aravis)

        # Most of these are slight guess work, for now this is not an issue.
        # Might want to validate whether the spec's match the expected type.
        self.feature_map = {
            "DeviceReset": MagicMock(spec=Aravis.GcCommand),
            "DeviceSerialNumber": MagicMock(spec=Aravis.GcRegisterNode),
            "DeviceVersion": MagicMock(spec=Aravis.GcRegisterNode),
            "AcquisitionFrameRateAuto": MagicMock(spec=Aravis.GcEnumeration),
            "AcquisitionFrameRateEnabled": MagicMock(spec=Aravis.GcBoolean),
            "ExposureMode": MagicMock(spec=Aravis.GcEnumeration),
            "ChunkSelector": MagicMock(spec=Aravis.GcEnumeration),
            "ChunkEnable": MagicMock(spec=Aravis.GcBoolean),
            "DeviceTemperature": MagicMock(spec=Aravis.GcIntegerNode),
            "TransmitFailureCount": MagicMock(spec=Aravis.GcIntegerNode),
            "BlackLevel": MagicMock(spec=Aravis.GcIntegerNode),
            "DeviceIndicatorMode": MagicMock(spec=Aravis.GcEnumeration),
            "DeviceModelName": MagicMock(spec=Aravis.GcRegisterNode),
            "DeviceID": MagicMock(spec=Aravis.GcRegisterNode),
            "DeviceFirmwareVersion": MagicMock(spec=Aravis.GcRegisterNode),
            "DeviceVendorName": MagicMock(spec=Aravis.GcRegisterNode),
            "pgrPowerSupplyVoltage": MagicMock(spec=Aravis.GcFloatNode),
            "pgrPowerSupplyCurrent": MagicMock(spec=Aravis.GcFloatNode),
            "pgrSensorDescription": MagicMock(spec=Aravis.GcRegisterDescriptionNode),
        }

    def tearDown(self) -> None:
        qmi.stop()
        logging.getLogger("qmi.instruments.ptgrey.blackfly_aravis").setLevel(logging.NOTSET)

    def test_open_ser(self):
        """Test whether a camera is found using a serial number."""
        instr = self._create_instr_ser()
        self.aravis.get_n_devices = MagicMock(return_value=1)
        self.aravis.get_device_serial_nbr = MagicMock(return_value="SerialNumber")
        self.aravis.get_device_address = MagicMock(return_value="10.10.10.13")
        instr.open()
        self.aravis.Camera.new.assert_called_once_with(self.aravis.get_device_id(1))
        instr.close()

    def test_open_ser_multiple(self):
        """Test whether the correct serial number camera is selected from a list of cameras."""
        instr = self._create_instr_ser()
        self.aravis.get_n_devices = MagicMock(return_value=3)
        self.aravis.get_device_serial_nbr = MagicMock(
            side_effect=[
                "BadSerial1",
                "BadSerial2",
                "SerialNumber",
            ],
        )
        self.aravis.get_device_address = MagicMock(return_value="10.10.10.13")
        instr.open()
        self.aravis.Camera.new.assert_called_once_with(self.aravis.get_device_id(1))
        instr.close()

    def test_open_ip(self):
        """Test whether a camera is found using a ip address."""
        instr = self._create_instr_ip()
        self.aravis.get_n_devices = MagicMock(return_value=1)
        self.aravis.get_device_serial_nbr = MagicMock(return_value="BadSerial")
        self.aravis.get_device_address = MagicMock(return_value="10.10.10.10")
        instr.open()
        self.aravis.Camera.new.assert_called_once_with(self.aravis.get_device_id(1))
        instr.close()

    def test_open_ip_multiple(self):
        """Test whether the correct ip address camera is selected from a list of cameras."""
        instr = self._create_instr_ip()
        self.aravis.get_n_devices = MagicMock(return_value=3)
        self.aravis.get_device_serial_nbr = MagicMock(return_value="BadSerial")
        self.aravis.get_device_address = MagicMock(
            side_effect=[
                "10.10.10.11",
                "10.10.10.12",
                "10.10.10.10",
            ],
        )
        instr.open()
        self.aravis.Camera.new.assert_called_once_with(self.aravis.get_device_id(1))
        instr.close()

    def test_open_no_cam(self):
        """Test whether an exception is raised when no cameras are found."""
        instr = self._create_instr_ser()
        self.aravis.get_n_devices = MagicMock(return_value=1)
        self.aravis.get_device_serial_nbr = MagicMock(return_value="BadSerial")
        self.aravis.get_device_address = MagicMock(return_value="10.10.10.13")
        with self.assertRaises(QMI_InstrumentException):
            instr.open()

    def test_open_duplicate(self):
        """Test whether an exception is raised duplicate cameras are found."""
        instr = self._create_instr_ser()
        self.aravis.get_n_devices = MagicMock(return_value=3)
        self.aravis.get_device_serial_nbr = MagicMock(
            side_effect=[
                "SerialNumber",
                "SerialNumber",
                "SerialNumber",
            ],
        )
        self.aravis.get_device_address = MagicMock(return_value="10.10.10.13")
        with self.assertRaises(QMI_InstrumentException):
            instr.open()

    def test_close(self):
        """Test the close function of a camera."""
        instr = self._create_instr_ser()
        c = self.aravis.Camera
        self.aravis.get_n_devices = MagicMock(return_value=1)
        self.aravis.get_device_serial_nbr = MagicMock(return_value="SerialNumber")
        self.aravis.get_device_address = MagicMock(return_value="10.10.10.13")
        self.aravis.Camera.new = MagicMock(return_value=c)
        instr.open()
        instr.close()
        c.stop_acquisition.assert_called_with()

    def _instr_create_camera(self):
        """Create a camera and allow it to be found when PtGreyBlackFly.open
        is called. PtGreyBlackFly containing the correct serial number or ip address."""
        camera = MagicMock(spec=Aravis.Camera)
        self.aravis.get_device_serial_nbr = MagicMock(return_value="SerialNumber")
        self.aravis.get_device_address = MagicMock(return_value="10.10.10.13")
        self.aravis.get_device_serial_nbr = MagicMock(return_value="SerialNumber")
        self.aravis.Camera.new = MagicMock(return_value=camera)
        return camera

    def _instr_open_obj(self):
        """Create PtGreyBlackFly object and a camera with the same SerialNumber.
        Open the object."""
        instr = self._create_instr_obj()
        camera = self._instr_create_camera()
        instr.open()
        return instr, camera

    def _instr_open(self):
        """Create PtGreyBlackFly instrument and a camera with the same SerialNumber.
        Open the instrument."""
        instr = self._create_instr_ser()
        camera = self._instr_create_camera()
        instr.open()
        return instr, camera

    def _create_device(self):
        """Mock a Aravis.GvDevice and attach the self._get_feature function
        to its .get_feature side_effect. This results in the same functionality
        that is expected from the device library. Features can be collected
        with self._get_feature by string, and by string they'll be identical."""
        device = MagicMock(Aravis.GvDevice)
        device.get_feature = MagicMock(side_effect=self._get_feature)
        return device

    def test_reset(self):
        """Test PtGreyBlackFly.reset()."""
        instr, camera = self._instr_open()
        device = self._create_device()
        camera.get_device = MagicMock(return_value=device)
        with patch("qmi.instruments.ptgrey.blackfly_aravis.time.sleep", MagicMock()):
            instr.reset()
            self._get_feature("DeviceReset").execute.assert_called_once_with()
            instr.close()

    def test_get_idn(self):
        """Test PtGreyBlackFly.get_idn()."""
        instr, camera = self._instr_open()
        device = self._create_device()

        camera.get_device = MagicMock(return_value=device)
        i = instr.get_idn()
        self.assertEqual(i.vendor, camera.get_vendor_name())
        self.assertEqual(i.model, camera.get_model_name())
        self.assertEqual(
            i.serial, self._get_feature("DeviceSerialNumber").get_value_as_string()
        )
        self.assertEqual(
            i.version, self._get_feature("DeviceVersion").get_value_as_string()
        )
        instr.close()

    def test_get_ip_address(self):
        """Test PtGreyBlackFly.get_ip_address()."""
        instr, camera = self._instr_open()
        device = self._create_device()
        address = MagicMock(spec=Aravis.Address)
        inetaddress = MagicMock(spec=Aravis.InetAddress)
        device.get_device_address = MagicMock(return_value=address)
        camera.get_device = MagicMock(return_value=device)
        address.get_address = MagicMock(return_value=inetaddress)
        i = instr.get_ip_address()
        self.assertEqual(i, address.get_address().to_string())
        instr.close()

    def test_get_device_info(self):
        """Test PtGreyBlackFly.get_device_info()."""
        instr, camera = self._instr_open()
        device = self._create_device()
        camera.get_device = MagicMock(return_value=device)
        i = instr.get_device_info()
        for key in i:
            self.assertEqual(
                i[key],
                self._get_feature(key).get_value_as_string(),
                msg=f"{key} values are not equal",
            )
        instr.close()

    def test_get_device_info_gdouble(self):
        """Test PtGreyBlackFly.get_device_info() where DeviceVendorName is a gdouble.
        This is a mocked, the real scenario likely does not have DeviceVendorName as a gdouble.
        """
        instr, camera = self._instr_open()
        device = self._create_device()
        feature = self._get_feature("DeviceVendorName")
        v = MagicMock()
        v.name = "gdouble"
        camera.get_device = MagicMock(return_value=device)
        feature.get_value_type = MagicMock(return_value=v)
        i = instr.get_device_info()
        self.assertEqual(i["DeviceVendorName"], feature.get_value())
        instr.close()

    def test_get_frame_rate(self):
        """Test PtGreyBlackFly.get_frame_rate()."""
        instr, camera = self._instr_open()
        f = instr.get_frame_rate()
        self.assertEqual(f, camera.get_frame_rate())
        instr.close()

    def test_get_frame_rate_auto(self):
        """Test PtGreyBlackFly.get_frame_rate_auto()."""
        instr, camera = self._instr_open()
        device = self._create_device()
        camera.get_device = MagicMock(return_value=device)
        f = instr.get_frame_rate_auto()
        self.assertEqual(
            f, self._get_feature("AcquisitionFrameRateAuto").get_string_value()
        )
        instr.close()

    def test_set_frame_rate(self):
        """Test PtGreyBlackFly.set_frame_rate()."""
        instr, camera = self._instr_open()
        device = self._create_device()
        camera.get_device = MagicMock(return_value=device)
        r = MagicMock()
        a = MagicMock()
        instr.set_frame_rate(r, a)
        self._get_feature(
            "AcquisitionFrameRateEnabled"
        ).set_value.assert_called_once_with(True)
        self._get_feature(
            "AcquisitionFrameRateAuto"
        ).set_string_value.assert_called_once_with(a)
        camera.set_frame_rate(r)
        instr.close()

    def test_get_gain(self):
        """Test PtGreyBlackFly.get_gain()."""
        instr, camera = self._instr_open()
        g = instr.get_gain()
        self.assertEqual(g, camera.get_gain())
        instr.close()

    def test_get_gain_auto(self):
        """Test PtGreyBlackFly.get_gain_auto()."""
        instr, camera = self._instr_open()
        g = instr.get_gain_auto()
        self.assertEqual(g, self.aravis.Auto.to_string(camera.get_gain_auto()))
        instr.close()

    def test_set_gain(self):
        """Test PtGreyBlackFly.set_gain()."""
        instr, camera = self._instr_open()
        g = MagicMock()
        a = MagicMock()
        instr.set_gain(g, a)
        camera.set_gain_auto.assert_called_once_with(self.aravis.Auto.from_string(a))
        camera.set_gain.assert_called_once_with(g)
        instr.close()

    def test_get_exposure_mode(self):
        """Test PtGreyBlackFly.get_exposure_mode()."""
        instr, camera = self._instr_open()
        device = self._create_device()
        camera.get_device = MagicMock(return_value=device)
        e = instr.get_exposure_mode()
        self.assertEqual(e, self._get_feature("ExposureMode").get_string_value())
        instr.close()

    def test_set_exposure_mode(self):
        """Test PtGreyBlackFly.set_exposure_mode()."""
        instr, camera = self._instr_open()
        device = self._create_device()

        camera.get_device = MagicMock(return_value=device)
        m = MagicMock()
        instr.set_exposure_mode(m)
        self._get_feature("ExposureMode").set_string_value.assert_called_once_with(m)
        instr.close()

    def test_get_exposure_time(self):
        """Test PtGreyBlackFly.get_exposure_time()."""
        instr, camera = self._instr_open()
        e = instr.get_exposure_time()
        self.assertEqual(e, camera.get_exposure_time())
        instr.close()

    def test_get_exposure_time_auto(self):
        """Test PtGreyBlackFly.get_exposure_time_auto()."""
        instr, camera = self._instr_open()
        e = instr.get_exposure_time_auto()
        self.assertEqual(e, self.aravis.Auto.to_string(camera.get_exposure_time_auto()))
        instr.close()

    def test_set_exposure_time(self):
        """Test PtGreyBlackFly.set_exposure_time()."""
        instr, camera = self._instr_open()
        device = self._create_device()
        camera.get_device = MagicMock(return_value=device)
        e = MagicMock()
        a = MagicMock()
        instr.set_exposure_time(e, a)
        camera.set_exposure_time_auto.assert_called_once_with(
            self.aravis.Auto.from_string(a)
        )
        camera.set_exposure_time(e)
        instr.close()

    def test_get_pixel_format(self):
        """Test PtGreyBlackFly.get_pixel_format()."""
        instr, camera = self._instr_open()
        with patch.object(
            camera,
            "get_pixel_format",
            side_effect=[
                self.aravis.PIXEL_FORMAT_MONO_8,
                self.aravis.PIXEL_FORMAT_MONO_16,
                self.aravis.PIXEL_FORMAT_MONO_12_PACKED,
            ],
        ):
            p1 = instr.get_pixel_format()
            p2 = instr.get_pixel_format()
            p3 = instr.get_pixel_format()
            self.assertEqual(p1, "Mono8")
            self.assertEqual(p2, "Mono16")
            self.assertEqual(p3, "Mono12Packed")
        instr.close()

    def test_get_pixel_format_invalid(self):
        """Test PtGreyBlackFly.get_pixel_format_invalid()."""
        instr, camera = self._instr_open()
        with patch.object(
            camera,
            "get_pixel_format",
            return_value=0xBEEFCAFE,
        ):
            with self.assertRaises(ValueError):
                instr.get_pixel_format()
        instr.close()

    def test_set_pixel_format(self):
        """Test PtGreyBlackFly.set_pixel_format()."""
        instr, camera = self._instr_open()
        instr.set_pixel_format("Mono8")
        camera.set_pixel_format.assert_called_once_with(self.aravis.PIXEL_FORMAT_MONO_8)
        camera.set_pixel_format.reset_mock()
        instr.set_pixel_format("Mono16")
        camera.set_pixel_format.assert_called_once_with(
            self.aravis.PIXEL_FORMAT_MONO_16
        )
        camera.set_pixel_format.reset_mock()
        instr.set_pixel_format("Mono12Packed")
        camera.set_pixel_format.assert_called_once_with(
            self.aravis.PIXEL_FORMAT_MONO_12_PACKED
        )
        camera.set_pixel_format.reset_mock()
        instr.close()

    def test_set_pixel_format_invalid(self):
        """Test PtGreyBlackFly.set_pixel_format_invalid()."""
        instr, _ = self._instr_open()
        with self.assertRaises(ValueError):
            instr.set_pixel_format("INVALID!!!!")
        instr.close()

    def test_get_image_size(self):
        """Test PtGreyBlackFly.get_image_size()."""
        instr, camera = self._instr_open()
        r = Region
        with patch.object(
            camera,
            "get_region",
            return_value=r,
        ):
            w, h = instr.get_image_size()
            self.assertEqual(w, r.width)
            self.assertEqual(h, r.height)
        instr.close()

    def test_set_image_size(self):
        """Test PtGreyBlackFly.set_image_size()."""
        instr, camera = self._instr_open()
        r = Region
        with patch.object(
            camera,
            "get_region",
            return_value=r,
        ):
            w = MagicMock()
            h = MagicMock()
            instr.set_image_size(w, h)
            camera.set_region.assert_called_once_with(r.x, r.y, w, h)
        instr.close()

    def test_get_image_offset(self):
        """Test PtGreyBlackFly.get_image_offset()."""
        instr, camera = self._instr_open()
        r = Region
        with patch.object(
            camera,
            "get_region",
            return_value=r,
        ):
            x, y = instr.get_image_offset()
            self.assertEqual(x, r.x)
            self.assertEqual(y, r.y)
        instr.close()

    def test_set_image_offset(self):
        """Test PtGreyBlackFly.set_image_offset()."""
        instr, camera = self._instr_open()
        r = Region
        with patch.object(
            camera,
            "get_region",
            return_value=r,
        ):
            x = MagicMock()
            y = MagicMock()
            instr.set_image_offset(x, y)
            camera.set_region.assert_called_once_with(x, y, r.width, r.height)
        instr.close()

    def test_get_chunk_mode(self):
        """Test PtGreyBlackFly.get_chunk_mode()."""
        instr, camera = self._instr_open()
        device = self._create_device()
        feature = self._get_feature("ChunkSelector")
        s = MagicMock()
        camera.get_device = MagicMock(return_value=device)
        feature.get_available_string_values = MagicMock(return_value=[s])
        a, t = instr.get_chunk_mode()
        self.assertEqual(a, camera.get_chunk_mode())
        feature.set_string_value.assert_called_once_with(s)
        self.assertEqual(t[s], self._get_feature("ChunkEnable").get_value())
        instr.close()

    def test_set_chunk_mode(self):
        """Test PtGreyBlackFly.set_chunk_mode()."""
        instr, camera = self._instr_open()
        device = self._create_device()
        s = MagicMock()
        feature = self._get_feature("ChunkSelector")
        camera.get_device = MagicMock(return_value=device)
        feature.get_available_string_values = MagicMock(return_value=[s])
        v = MagicMock()
        instr.set_chunk_mode(v)
        device.write_register.assert_called_once_with(0x0D24, 1)
        feature.set_string_value.assert_called_once_with(s)
        self._get_feature("ChunkEnable").set_value.assert_called_once_with(True)
        instr.close()

    def test_get_device_temperature(self):
        """Test PtGreyBlackFly.get_device_temperature()."""
        instr, camera = self._instr_open()
        device = self._create_device()
        camera.get_device = MagicMock(return_value=device)
        t = instr.get_device_temperature()
        self.assertEqual(t, self._get_feature("DeviceTemperature").get_value())
        instr.close()

    def test_get_transmit_failure_count(self):
        """Test PtGreyBlackFly.get_transmit_failure_count()."""
        instr, camera = self._instr_open()
        device = self._create_device()
        camera.get_device = MagicMock(return_value=device)
        count = instr.get_transmit_failure_count()
        self.assertEqual(count, self._get_feature("TransmitFailureCount").get_value())
        instr.close()

    def test_get_black_level(self):
        """Test PtGreyBlackFly.get_black_level()."""
        instr, camera = self._instr_open()
        device = self._create_device()
        camera.get_device = MagicMock(return_value=device)
        t = instr.get_black_level()
        self.assertEqual(t, self._get_feature("BlackLevel").get_value())
        instr.close()

    def test_set_black_level(self):
        """Test PtGreyBlackFly.set_black_level()."""
        instr, camera = self._instr_open()
        device = self._create_device()
        camera.get_device = MagicMock(return_value=device)
        l = MagicMock()
        instr.set_black_level(l)
        self._get_feature("BlackLevel").set_value.assert_called_once_with(l)
        instr.close()

    def test_set_indicator_mode_enable(self):
        """Test PtGreyBlackFly.set_indicator_mode_enable()."""
        instr, camera = self._instr_open()
        device = self._create_device()
        camera.get_device = MagicMock(return_value=device)
        instr.set_indicator_mode(True)
        self._get_feature(
            "DeviceIndicatorMode"
        ).set_string_value.assert_called_once_with("Active")
        instr.close()

    def test_set_indicator_mode_disable(self):
        """Test PtGreyBlackFly.set_indicator_mode_disable()."""
        instr, camera = self._instr_open()
        device = self._create_device()
        camera.get_device = MagicMock(return_value=device)
        instr.set_indicator_mode(False)
        self._get_feature(
            "DeviceIndicatorMode"
        ).set_string_value.assert_called_once_with("Inactive")
        instr.close()

    def test_is_streaming(self):
        """Test PtGreyBlackFly.is_streaming()."""
        instr, camera = self._instr_open()
        v = instr.is_streaming()
        self.assertEqual(v, False)
        camera.get_payload = MagicMock(return_value=1)
        instr.start_acquisition(1)
        v = instr.is_streaming()
        self.assertEqual(v, True)
        instr.close()

    def test_start_acquisition(self):
        """Test PtGreyBlackFly.start_acquisition()."""
        instr, camera = self._instr_open_obj()
        s = self.aravis.Stream
        c = self.aravis.ChunkParser
        b = self.aravis.Buffer
        camera.get_payload = MagicMock(return_value=1)
        camera.create_stream = MagicMock(return_value=s)
        camera.create_chunk_parser = MagicMock(return_value=c)
        self.aravis.Buffer.new = MagicMock(return_value=b)
        with patch("qmi.instruments.ptgrey.blackfly_aravis.min", return_value=1):
            instr.start_acquisition(None)
            camera.set_acquisition_mode.assert_called_once_with(
                self.aravis.AcquisitionMode.CONTINUOUS
            )
            s.connect.assert_called_once_with("new-buffer", instr._new_image_cb)
            s.set_emit_signals.assert_called_once_with(False)
            s.push_buffer.assert_called_once_with(b)
            camera.start_acquisition.assert_called_once_with()
        instr.close()

    def test_start_acquisition_cb_enable(self):
        """Test PtGreyBlackFly.start_acquisition() with callback enabled"""
        instr, camera = self._instr_open_obj()
        s = self.aravis.Stream
        camera.get_payload = MagicMock(return_value=1)
        camera.create_stream = MagicMock(return_value=s)
        instr.register_image_callback(MagicMock())
        instr.start_acquisition(None)
        s.set_emit_signals.assert_called_once_with(True)
        instr.close()

    def test_start_acquisition_continuous(self):
        """Test PtGreyBlackFly.start_acquisition_continues()"""
        instr, camera = self._instr_open()
        camera.get_payload = MagicMock(return_value=1)
        instr.start_acquisition(None)
        camera.set_acquisition_mode.assert_called_once_with(
            self.aravis.AcquisitionMode.CONTINUOUS
        )
        instr.close()

    def test_start_acquisition_single(self):
        """Test PtGreyBlackFly.start_acquisition_single()"""
        instr, camera = self._instr_open()
        camera.get_payload = MagicMock(return_value=1)
        instr.start_acquisition(1)
        camera.set_acquisition_mode.assert_called_once_with(
            self.aravis.AcquisitionMode.SINGLE_FRAME
        )
        instr.close()

    def test_start_acquisition_multi(self):
        """Test PtGreyBlackFly.start_acquisition_multi()"""
        instr, camera = self._instr_open()
        camera.get_payload = MagicMock(return_value=1)
        instr.start_acquisition(65535)
        camera.set_acquisition_mode.assert_called_once_with(
            self.aravis.AcquisitionMode.MULTI_FRAME
        )
        camera.set_frame_count(65535)
        instr.close()

    def test_start_acquisition_invalid(self):
        """Test PtGreyBlackFly.start_acquisition() with invalid input."""
        instr, camera = self._instr_open()
        with self.assertRaises(QMI_UsageException):
            instr.start_acquisition(65537)
        instr.close()

    def test_stop_acquisition(self):
        """Test PtGreyBlackFly.stop_acquisition()."""
        instr, camera = self._instr_open()
        s = self.aravis.Stream
        camera.get_payload = MagicMock(return_value=1)
        camera.create_stream = MagicMock(return_value=s)
        instr.start_acquisition(None)
        s.set_emit_signals.reset_mock()
        instr.stop_acquisition()
        s.set_emit_signals.assert_called_once_with(False)
        camera.stop_acquisition.assert_called_once_with()
        instr.close()

    def test_get_next_image(self):
        """Test PtGreyBlackFly.get_next_image()."""
        instr, _ = self._instr_open_obj()
        s = self.aravis.Stream
        b = self.aravis.Buffer
        instr._stream = s
        instr._convert_image = MagicMock()
        s.pop_buffer = MagicMock(return_value=b)
        b.get_status = MagicMock(return_value=self.aravis.BufferStatus.SUCCESS)
        instr.get_next_image(None)
        s.pop_buffer.assert_called_once_with()
        instr._convert_image.assert_called_once_with(b)
        s.push_buffer.assert_called_once_with(b)
        instr.close()

    def test_get_next_image_timeout(self):
        """Test PtGreyBlackFly.get_next_image_timeout()."""
        instr, _ = self._instr_open_obj()
        s = self.aravis.Stream
        b = self.aravis.Buffer
        instr._stream = s
        instr._convert_image = MagicMock()
        s.timeout_pop_buffer = MagicMock(return_value=b)
        b.get_status = MagicMock(return_value=self.aravis.BufferStatus.SUCCESS)
        instr.get_next_image(100)  # Timeout is in us
        s.timeout_pop_buffer.assert_called_once_with(int(1000000 * 100))
        s.push_buffer.assert_called_once_with(b)
        instr.close()

    def test_get_next_image_no_stream(self):
        """Test PtGreyBlackFly.get_next_image_no_stream()."""
        instr, _ = self._instr_open_obj()
        with self.assertRaises(QMI_InstrumentException):
            instr.get_next_image(None)
        instr.close()

    def test_get_next_image_no_buffer(self):
        """Test PtGreyBlackFly.get_next_image_no_buffer()."""
        instr, _ = self._instr_open_obj()
        s = self.aravis.Stream
        instr._stream = s
        instr._convert_image = MagicMock()
        s.pop_buffer = MagicMock(return_value=None)
        with self.assertRaises(QMI_TimeoutException):
            instr.get_next_image(None)
        instr.close()

    def test_get_next_image_bad_status(self):
        """Test PtGreyBlackFly.get_next_image_bad_status()."""
        instr, _ = self._instr_open_obj()
        s = self.aravis.Stream
        b = self.aravis.Buffer
        instr._stream = s
        instr._convert_image = MagicMock()
        s.pop_buffer = MagicMock(return_value=b)
        b.get_status = MagicMock(return_value=self.aravis.BufferStatus.FAILURE)
        with self.assertRaises(QMI_InstrumentException):
            instr.get_next_image(None)
        instr.close()

    def test_register_image_callback(self):
        """Test PtGreyBlackFly.register_image_callback()."""
        instr, _ = self._instr_open_obj()
        s = self.aravis.Stream
        instr._stream = s
        c = MagicMock()
        instr.register_image_callback(c)
        self.assertEqual(instr._image_callbacks, [c])
        s.set_emit_signals.assert_called_once_with(True)
        instr.close()

    def test_register_image_callback_no_stream(self):
        """Test PtGreyBlackFly.register_image_callback_no_stream()."""
        instr, _ = self._instr_open_obj()
        s = self.aravis.Stream
        c = MagicMock()
        instr.register_image_callback(c)
        self.assertEqual(instr._image_callbacks, [c])
        s.set_emit_signals.assert_not_called()
        instr.close()

    def test_unregister_image_callback(self):
        """Test PtGreyBlackFly.unregister_image_callback()."""
        instr, _ = self._instr_open_obj()
        c = MagicMock()
        instr.register_image_callback(c)
        self.assertEqual(instr._image_callbacks, [c])
        instr.unregister_image_callback(c)
        self.assertEqual(instr._image_callbacks, [])
        instr.close()

    def test_unregister_image_callback_invalid(self):
        """Test PtGreyBlackFly.unregister_image_callback() invalid input."""
        instr, _ = self._instr_open_obj()
        c = MagicMock()
        with self.assertRaises(ValueError):
            instr.unregister_image_callback(c)
        instr.close()

    def test_new_image_cb(self):
        """Test PtGreyBlackFly.new_image_cb()."""
        instr, _ = self._instr_open_obj()
        b = self.aravis.Buffer
        s = self.aravis.Stream
        c = MagicMock()
        m = MagicMock()
        s.try_pop_buffer = MagicMock(return_value=b)
        b.get_status = MagicMock(return_value=self.aravis.BufferStatus.SUCCESS)
        instr._image_callbacks = [c]
        instr._convert_image = m
        instr._new_image_cb(s)
        c.assert_called_once_with(instr._convert_image(b))
        instr.close()

    def test_new_image_cb_no_buffer(self):
        """Test PtGreyBlackFly.new_image_cb() without buffer."""
        instr, _ = self._instr_open_obj()
        b = self.aravis.Buffer
        s = self.aravis.Stream
        s.try_pop_buffer = MagicMock(return_value=None)
        instr._new_image_cb(s)
        b.get_status.assert_not_called()
        instr.close()

    def test_new_image_cb_buf_no_suc(self):
        """Test PtGreyBlackFly.new_image_cb() without buffer success."""
        instr, _ = self._instr_open_obj()
        b = self.aravis.Buffer
        s = self.aravis.Stream
        m = MagicMock()
        s.try_pop_buffer = MagicMock(return_value=b)
        b.get_status = MagicMock(return_value=self.aravis.BufferStatus.FAILURE)
        instr._convert_image = m
        instr._new_image_cb(s)
        instr._convert_image.assert_not_called()
        instr.close()

    def test_new_image_cb_exception(self):
        """Test PtGreyBlackFly.new_image_cb() with callbakc exception."""
        instr, _ = self._instr_open_obj()
        b = self.aravis.Buffer
        s = self.aravis.Stream
        c = MagicMock(side_effect=Exception)
        m = MagicMock()
        s.try_pop_buffer = MagicMock(return_value=b)
        b.get_status = MagicMock(return_value=self.aravis.BufferStatus.SUCCESS)
        instr._image_callbacks = [c]
        instr._convert_image = m
        with patch(
            "qmi.instruments.ptgrey.blackfly_aravis._logger", MagicMock()
        ):  # remove exception log during unittest run.
            instr._new_image_cb(s)
            c.assert_called_once_with(instr._convert_image(b))
        instr.close()

    def test_convert_image(self):
        """Test PtGreyBlackFly.convert_image()."""
        instr, _ = self._instr_open_obj()
        parser = self.aravis.ChunkParser
        gain = MagicMock()
        black_level = MagicMock()
        exposure_time = MagicMock()
        buffer = self.aravis.Buffer
        width = 2
        height = 1

        def side_effect_get_float_value(buf, str):
            """Return a mocked float value depending on the provided string."""
            if buf is buffer:
                if str == "ChunkGain":
                    return gain
                if str == "ChunkBlackLevel":
                    return black_level
                if str == "ChunkExposureTime":
                    return exposure_time

        instr._chunk_parser = parser
        buffer.get_status = MagicMock(return_value=self.aravis.BufferStatus.SUCCESS)
        buffer.get_image_pixel_format = MagicMock(
            return_value=self.aravis.PIXEL_FORMAT_MONO_8
        )
        buffer.get_data = MagicMock(return_value=b"/x1/x2/x3/x4/x5")
        buffer.get_image_width = MagicMock(return_value=width)
        buffer.get_image_height = MagicMock(return_value=height)
        buffer.get_payload_type = MagicMock(
            return_value=self.aravis.BufferPayloadType.EXTENDED_CHUNK_DATA
        )
        parser.get_float_value = MagicMock(side_effect=side_effect_get_float_value)
        info = instr._convert_image(buffer)
        # Manipulate the buffer data equal to _convert_image() MONO_8.
        data = np.frombuffer(buffer.get_data(), dtype=np.uint8, count=(height * width))
        data = data.reshape((height, width))
        # Image info should be equal to that of the buffer.
        self.assertEqual(info.width, buffer.get_image_width())
        self.assertEqual(info.height, buffer.get_image_height())
        self.assertEqual(info.offset_x, buffer.get_image_x())
        self.assertEqual(info.offset_y, buffer.get_image_y())
        self.assertEqual(info.pixel_format, buffer.get_image_pixel_format())
        self.assertEqual(info.frame_id, buffer.get_frame_id())
        self.assertEqual(info.image_id, buffer.get_frame_id())
        self.assertEqual(info.timestamp, 1.0e-9 * buffer.get_timestamp())
        self.assertEqual(info.gain, gain)
        self.assertEqual(info.black_level, black_level)
        self.assertEqual(info.exposure_time, exposure_time)
        np.testing.assert_array_equal(info.data, data)
        instr.close()

    def test_convert_image_mono_16(self):
        """Test PtGreyBlackFly.convert_image() with MONO_16 format."""
        instr, _ = self._instr_open_obj()
        buffer = self.aravis.Buffer
        width = 2
        height = 1

        instr._chunk_parser = MagicMock()
        buffer.get_status = MagicMock(return_value=self.aravis.BufferStatus.SUCCESS)
        buffer.get_image_pixel_format = MagicMock(
            return_value=self.aravis.PIXEL_FORMAT_MONO_16
        )
        buffer.get_data = MagicMock(return_value=b"/x1/x2/x3/x4/x5")
        buffer.get_image_width = MagicMock(return_value=width)
        buffer.get_image_height = MagicMock(return_value=height)
        i = instr._convert_image(buffer)
        # Manipulate the buffer data equal to _convert_image() MONO_16.
        data = np.frombuffer(buffer.get_data(), dtype="<H", count=height * width)
        data = data.reshape((height, width))
        np.testing.assert_array_equal(i.data, data)
        instr.close()

    def test_convert_image_mono_12_packed(self):
        """Test PtGreyBlackFly.convert_image() with MONO_12_packed format."""
        instr, _ = self._instr_open_obj()
        buffer = self.aravis.Buffer
        width = 2
        height = 1

        instr._chunk_parser = MagicMock()
        buffer.get_status = MagicMock(return_value=self.aravis.BufferStatus.SUCCESS)
        buffer.get_image_pixel_format = MagicMock(
            return_value=self.aravis.PIXEL_FORMAT_MONO_12_PACKED
        )
        buffer.get_data = MagicMock(return_value=b"/x1/x2/x3/x4/x5")
        buffer.get_image_width = MagicMock(return_value=width)
        buffer.get_image_height = MagicMock(return_value=height)
        i = instr._convert_image(buffer)
        # Manipulate the buffer data equal to _convert_image() MONO_12_PACKED.
        tmp = np.frombuffer(
            buffer.get_data(), dtype=np.uint8, count=(height * width * 3 + 1) // 2
        )
        tmp = tmp.astype(np.uint16)
        data = np.zeros(2 * 1, dtype=np.uint16)
        data[::2] = ((tmp[1::3] & 0x0F) << 8) | tmp[0::3]
        data[1::2] = (tmp[2::3] << 4) | ((tmp[1::3] & 0xF0) >> 4)
        data = data.reshape((height, width))
        np.testing.assert_array_equal(i.data, data)
        instr.close()

    def test_convert_image_mono_invalid_pf(self):
        """Test PtGreyBlackFly.convert_image() with invalid format"""
        instr, _ = self._instr_open_obj()
        buffer = self.aravis.Buffer

        buffer.get_status = MagicMock(return_value=self.aravis.BufferStatus.SUCCESS)
        buffer.get_image_pixel_format = MagicMock(return_value=0xBEEFCAFE)
        with self.assertRaises(QMI_InstrumentException):
            instr._convert_image(buffer)
        instr.close()

    def test_convert_image_no_chunk_parser(self):
        """Test PtGreyBlackFly.convert_image() without parser"""
        instr, _ = self._instr_open_obj()
        buffer = self.aravis.Buffer

        instr._chunk_parser = MagicMock()
        buffer.get_status = MagicMock(return_value=self.aravis.BufferStatus.SUCCESS)
        buffer.get_image_pixel_format = MagicMock(
            return_value=self.aravis.PIXEL_FORMAT_MONO_12_PACKED
        )
        buffer.get_data = MagicMock(return_value=b"/x1/x2/x3/x4/x5")
        buffer.get_image_width = MagicMock(return_value=1)
        buffer.get_image_height = MagicMock(return_value=2)
        i = instr._convert_image(buffer)
        self.assertEqual(i.gain, 0)
        self.assertEqual(i.black_level, 0)
        self.assertEqual(i.exposure_time, 0)
        instr.close()

    def test_convert_image_buf_stat_fail(self):
        """Test PtGreyBlackFly.convert_image() with buffer failure."""
        instr, _ = self._instr_open_obj()
        buffer = self.aravis.Buffer

        buffer.get_status = MagicMock(return_value=self.aravis.BufferStatus.FAILURE)
        with self.assertRaises(QMI_InstrumentException):
            instr._convert_image(buffer)
        instr.close()

    def test_convert_image_payload_chunk_data(self):
        """Test PtGreyBlackFly.convert_image() with invalid buffer payload type."""
        instr, _ = self._instr_open_obj()
        buffer = self.aravis.Buffer

        buffer.get_status = MagicMock(return_value=self.aravis.BufferStatus.SUCCESS)
        buffer.get_payload_type = MagicMock(
            return_value=self.aravis.BufferPayloadType.CHUNK_DATA
        )
        with self.assertRaises(QMI_InstrumentException):
            instr._convert_image(buffer)
        instr.close()

    def test_list_instrument(self):
        """Test PtGreyBlackFly.list_instruments()."""
        self.aravis.get_n_devices = MagicMock(return_value=1)
        d = Flir_Blackfly_Aravis.list_instruments()
        self.assertEqual(d[0].vendor_name, self.aravis.get_device_vendor(1))
        self.assertEqual(d[0].model_name, self.aravis.get_device_model(1))
        self.assertEqual(d[0].serial_number, self.aravis.get_device_serial_nbr(1))
        self.assertEqual(d[0].version, "")
        self.assertEqual(d[0].ip_address, self.aravis.get_device_address(1))

    def test_init_usage_exception(self):
        """Create a PtGeyBlackFlyAravis object usage exception due to invalid input."""
        with self.assertRaises(QMI_UsageException):
            Flir_Blackfly_Aravis(
                QMI_Context("context"),
                "name",
                None,
                None,
            )



"""
Instrument driver for the Measurement Computing USB-1808X DAQ.
"""

import logging
import typing

from typing import List, Optional

from qmi.core.context import QMI_Context
from qmi.core.exceptions import QMI_InstrumentException
from qmi.core.instrument import QMI_Instrument, QMI_InstrumentIdentification
from qmi.core.rpc import rpc_method

# Lazy import of the "uldaq" module. See the function _import_modules() below.
if typing.TYPE_CHECKING:
    import uldaq
else:
    uldaq = None


_logger = logging.getLogger(__name__)


def _import_modules() -> None:
    """Import the "uldaq" library.

    This import is done in a function, instead of at the top-level,
    to avoid an unnecessary dependency for programs that do not access
    the instrument directly.
    """
    global uldaq
    if uldaq is None:
        import uldaq  # pylint: disable=W0621


class MCC_USB1808X(QMI_Instrument):
    """Instrument driver for the Measurement Computing USB-1808X DAQ.

    This driver implements only a subset of the features of the USB-1808X.
    Only programmatic read/write access to DIO pins and analog input/output
    is supported. Advanced features such as timers, counters, triggering,
    scanning, waveforms etc. are not yet supported.
    """

    def __init__(self, context: QMI_Context, name: str, unique_id: str) -> None:
        super().__init__(context, name)
        self._unique_id = unique_id
        self._device = None  # type: Optional[uldaq.DaqDevice]
        self._dio_device = None  # type: Optional[uldaq.DioDevice]
        self._ai_device = None  # type: Optional[uldaq.AiDevice]
        self._ao_device = None  # type: Optional[uldaq.AoDevice]

        # Import the "uldaq" module.
        _import_modules()

    @staticmethod
    def list_instruments() -> List[str]:
        """Return a list of unique_ids of connected devices."""

        _import_modules()
        instruments = []
        device_descriptors = uldaq.get_daq_device_inventory(uldaq.InterfaceType.USB)
        for device_descriptor in device_descriptors:
            if device_descriptor.product_name == "USB-1808X":
                instruments.append(device_descriptor.unique_id)
        return instruments

    @rpc_method
    def open(self) -> None:
        self._check_is_closed()

        device = None

        device_descriptors = uldaq.get_daq_device_inventory(uldaq.InterfaceType.USB)
        for device_descriptor in device_descriptors:
            if (device_descriptor.product_name == "USB-1808X") and (
                device_descriptor.unique_id == self._unique_id
            ):
                device = uldaq.DaqDevice(device_descriptor)

        if device is None:
            raise QMI_InstrumentException(
                f"USB-1808X with unique_id {self._unique_id!r} not found"
            )

        try:
            _logger.debug(
                "%s: Connecting to USB-1808X device %s", self._name, self._unique_id
            )
            device.connect()

            try:
                dio_device = device.get_dio_device()
                ai_device = device.get_ai_device()
                ao_device = device.get_ao_device()
            except:
                dio_device = None
                ai_device = None
                ao_device = None
                device.disconnect()
                raise

        except:
            device.release()
            raise

        self._device = device
        self._dio_device = dio_device
        self._ai_device = ai_device
        self._ao_device = ao_device

        super().open()

    @rpc_method
    def close(self) -> None:
        self._check_is_open()
        assert self._device is not None
        _logger.debug(
            "%s: Closing connection to USB-1808X device %s", self._name, self._unique_id
        )
        self._dio_device = None
        self._ai_device = None
        self._ao_device = None
        self._device.disconnect()
        self._device.release()
        self._device = None
        super().close()

    @rpc_method
    def get_idn(self) -> QMI_InstrumentIdentification:
        """Read instrument type and version and return a QMI_InstrumentIdentification instance."""
        self._check_is_open()
        assert self._device is not None
        desc = self._device.get_descriptor()
        cfg = self._device.get_config()
        version = cfg.get_version(uldaq.DevVersionType.FW_MAIN)
        return QMI_InstrumentIdentification(
            vendor="Measurement Computing",
            model=desc.product_name,
            serial=desc.unique_id,
            version=version,
        )

    @rpc_method
    def get_dio_num_channels(self) -> int:
        """Return the number of digital input/output channels."""
        self._check_is_open()
        assert self._dio_device is not None
        info = self._dio_device.get_info()
        return info.get_port_info(uldaq.DigitalPortType.AUXPORT).number_of_bits

    @rpc_method
    def get_dio_direction(self) -> List[bool]:
        """Return the configured direction (input or output) for each digital channel.

        :return: List of directions, where directions[i] is True if
                 channel i is configured for output, or directions[i] is False
                 if channel i is configured for input.
        """
        self._check_is_open()
        assert self._dio_device is not None
        cfg = self._dio_device.get_config()
        directions = cfg.get_port_direction(uldaq.DigitalPortType.AUXPORT)
        return [(d == uldaq.DigitalDirection.OUTPUT) for d in directions]

    @rpc_method
    def set_dio_direction(self, channel: int, output: bool) -> None:
        """Configure an digital channel as input or output channel.

        :param channel: Digital channel index (0 .. nchan-1).
        :param output: True to configure the channel for output;
                       False to configure the channel for input.
        """
        self._check_is_open()
        assert self._dio_device is not None
        d = uldaq.DigitalDirection.OUTPUT if output else uldaq.DigitalDirection.INPUT
        self._dio_device.d_config_bit(uldaq.DigitalPortType.AUXPORT, channel, d)

    @rpc_method
    def get_dio_input_bit(self, channel: int) -> bool:
        """Read input levels of the digital channels.

        :param channel: Digital channel index (0 .. nchan-1).
        :return: True if the input signal is high, False if the signal is low.
        """
        self._check_is_open()
        assert self._dio_device is not None
        return self._dio_device.d_bit_in(uldaq.DigitalPortType.AUXPORT, channel)

    @rpc_method
    def set_dio_output_bit(self, channel: int, value: bool) -> None:
        """Set a digital output channel to the specified level.

        The digital channel must be configured in output mode
        (see set_dio_direction) before using this function.

        :param channel: Digital channel index (0 .. nchan-1).
        :param value: True to set the output high, False to set the output low.
        """
        self._check_is_open()
        assert self._dio_device is not None
        return self._dio_device.d_bit_out(uldaq.DigitalPortType.AUXPORT, channel, value)

    @rpc_method
    def get_ai_num_channels(self) -> int:
        """Return the number of analog input channels."""
        self._check_is_open()
        assert self._ai_device is not None
        return self._ai_device.get_info().get_num_chans()

    @rpc_method
    def get_ai_ranges(self) -> List[str]:
        """Return the supported analog input ranges.

        Ranges have names such as "UNI2VOLTS" for 0 .. 2 Volt
        or "BIP10VOLTS" for -10 .. +10 Volt..
        """
        self._check_is_open()
        assert self._ai_device is not None
        ranges = self._ai_device.get_info().get_ranges(uldaq.AiChanType.VOLTAGE)
        return [urange.name for urange in ranges]

    @rpc_method
    def get_ai_value(self, channel: int, input_mode: str, analog_range: str) -> float:
        """Read the input voltage of an analog input channel.

        :param channel: Analog input channel index (0 .. nchan-1).
        :param input_mode: Input mode, "SINGLE_ENDED" or "DIFFERENTIAL".
        :param analog_range: Name of the input range to use for the channel.
        :return: Input level in Volt.
        """
        self._check_is_open()
        assert self._ai_device is not None
        mod = uldaq.AiInputMode[input_mode]
        urange = uldaq.Range[analog_range]
        return self._ai_device.a_in(channel, mod, urange, uldaq.AInFlag.DEFAULT)

    @rpc_method
    def get_ao_num_channels(self) -> int:
        """Return the number of analog output channels."""
        self._check_is_open()
        assert self._ao_device is not None
        return self._ao_device.get_info().get_num_chans()

    @rpc_method
    def get_ao_ranges(self) -> List[str]:
        """Return the supported analog output ranges.

        Ranges have names such as "UNI2VOLTS" for 0 .. 2 Volt
        or "BIP10VOLTS" for -10 .. +10 Volt..
        """
        self._check_is_open()
        assert self._ao_device is not None
        ranges = self._ao_device.get_info().get_ranges()
        return [urange.name for urange in ranges]

    @rpc_method
    def set_ao_value(self, channel: int, analog_range: str, value: float) -> None:
        """Set an analog output channel to the specified voltage.

        :param channel: Analog output channel index (0 .. nchan-1).
        :param analog_range: Name of the output range to use for the channel.
        :param value: Output level in Volt.
        """
        self._check_is_open()
        assert self._ao_device is not None
        urange = uldaq.Range[analog_range]
        self._ao_device.a_out(channel, urange, uldaq.AOutFlag.DEFAULT, value)



"""Instrument driver for the AGC302 Active Vacuum Gauge Controller."""

from typing import Union

from qmi.core.context import QMI_Context
from qmi.core.exceptions import QMI_InstrumentException
from qmi.core.instrument import QMI_Instrument
from qmi.core.rpc import rpc_method
from qmi.core.transport import create_transport, QMI_Transport


class InstruTech_AGC302(QMI_Instrument):
    """Instrument driver for the AGC302 Active Vacuum Gauge Controller."""

    def __init__(self, context: QMI_Context, name: str, transport: Union[str, QMI_Transport], timeout: float = 1):
        """Initialize the pressure gauge Agc302 driver.

        Args:
            context: QMI context.
            name: Name for this instrument instance.
            transport:  Either a transport string (see create_transport) or a QMI_Transport.
            timeout: Maximum time [s] to wait before transport timeout occurs.
        """
        super().__init__(context, name)
        if isinstance(transport, str):
            self._transport = create_transport(transport)
        elif isinstance(transport, QMI_Transport):
            self._transport = transport
        self._timeout = timeout

    @rpc_method
    def open(self) -> None:
        self._transport.open()
        super().open()

    @rpc_method
    def close(self) -> None:
        super().close()
        self._transport.close()

    def _ask(self, command: str) -> str:
        super()._check_is_open()
        bytes_command = "#  {}\r".format(command).encode(encoding='ascii')
        self._transport.write(bytes_command)
        response = self._transport.read_until(b'\r', self._timeout)
        if len(response) != 13:
            raise QMI_InstrumentException("Got wrong response from device: {!r}.".format(response))
        if response.startswith(b'?') and b'SYNTX_ER' in response:
            # Response to an unknown command
            raise QMI_InstrumentException("Syntax Error in command: {}.".format(command))
        return response[4:-1].decode('ascii')

    def _set(self, command: str) -> None:
        response = self._ask(command)
        if 'PROGM_OK' not in response:
            raise QMI_InstrumentException(
                "Did not successfully program device is command: {} ok?, got response {}.".format(command, response))

    @rpc_method
    def read_gauge(self) -> float:
        """Read the current gauge pressure.

        Returns:
            The device will return a pressure value in units of TORR, MBAR, or PASCAL depending on the current
            settings. Use read_pressure_unit to obtain the current used unit.

        Raises:
            QMI_InstrumentException: When sensor is off or when a device does not exists.
        """
        response = self._ask("RD")

        if "1.10E+03" in response:
            raise QMI_InstrumentException("Sensor is off")
        elif "9.90E+09" in response:
            raise QMI_InstrumentException("Device does not exist")
        else:
            return float(response)

    @rpc_method
    def read_pressure_unit(self) -> str:
        """Read the current set pressure unit.

        Returns:
            The gauge reports the measured pressure in Torr, mbar or Pascal. The return values are in capital letters so
            either TORR, MBAR, or PASCAL.
        """
        response = self._ask("RU")
        unit = response.split()
        return unit[0]

    @rpc_method
    def set_pressure_unit(self, unit: str) -> None:
        """Set pressure unit for display and RD response.

        Args:
            unit (str): pressure unit to use. A single letter needs to be provided that is either T = Torr, M = mBar, or
            P = Pascal.

        Raises:
            ValueError: when provided pressure unit letter is not T, M, or P.
        """
        if unit not in ['T', 'M', 'P']:
            raise ValueError('Unit not T = Torr, M = mBar, P = Pascal.')
        self._set('SU{}'.format(unit))



"""
Edwards Vacuum, TIC controller.

The qmi.instruments.edwards package provides support for:
- Turbo & Instrument Controller (TIC) models D397-21-000/D397-22-000.
"""
# Alternative, QMI naming convention approved name
from qmi.instruments.edwards.turbo_instrument_controller import Edwards_TurboInstrumentController as\
    EdwardsVacuum_TIC, EdwardsVacuum_TIC_StatusResponse, EdwardsVacuum_TIC_PumpStateResponse,\
    EdwardsVacuum_TIC_PumpSpeedResponse, EdwardsVacuum_TIC_PumpPowerResponse,\
    EdwardsVacuum_TIC_GaugePressureResponse, EdwardsVacuum_TIC_StateResponse, EdwardsVacuum_TIC_AlertId,\
    EdwardsVacuum_TIC_ErrorCode, EdwardsVacuum_TIC_GaugeState, EdwardsVacuum_TIC_Priority,\
    EdwardsVacuum_TIC_PumpState, EdwardsVacuum_TIC_State, EdwardsVacuum_TIC_SystemOnOffSetupConfigResponse
